/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "HeliFMM.h"

#include <stdio.h>
#include "debug.h"

#include "HeliTest/Events.h"
#include "SharedThreadIDs.h"
#include "SystemDiagnostics/SystemDiagnostics.h"
#include "events/EventBroker.h"

#include "HeliTest/config.h"

using miosix::Lock;

namespace RogallinaBoard
{

// Max priority (4) for the Flight Mode Manager
HeliFMM::HeliFMM(Cutter& cutter) : FSM(&HeliFMM::bootedState), cutter(cutter)
{
    sEventBroker->subscribe(this, TOPIC_USER_INPUT);
    sEventBroker->subscribe(this, TOPIC_STATUS);
    sEventBroker->subscribe(this, TOPIC_FSM);

    memset(&status, 0, sizeof(HeliStatus));
}

HeliFMM::~HeliFMM() { sEventBroker->unsubscribe(this); }

HeliStatus HeliFMM::getStatus() { return status; }

void HeliFMM::bootedState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
        {
            status.state = HELISTATE_BOOTED;
            logStatus();
            TRACE("[FMM] Entering booted state.\n");
            break;
        }
        case EV_EXIT:
            break;
#ifndef IGNORE_LOGGER_NOT_WORKING
        case EV_LOGGER_ERROR:
            transition(&HeliFMM::failedState);
            status.failReason = FMMFAIL_LOGGER;
            break;
#endif
#ifndef IGNORE_SENSORS_NOT_WORKING
        case EV_SENSOR_INIT_ERROR:
            transition(&HeliFMM::failedState);
            status.failReason = FMMFAIL_SENSOR;
            break;
#endif
        case EV_BUTTON_LONG_PRESS:
            transition(&HeliFMM::readyState);
            break;
#ifdef DEBUG
        case EV_BUTTON_LONG_LONG_PRESS:
            transition(&HeliFMM::calibrationState);
            break;
#endif
    }
}

void HeliFMM::failedState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = HELISTATE_FAILED;
            logStatus();
            TRACE("[FMM] Entering failed state.\n");
            break;
        case EV_EXIT:
            break;
    }
}

void HeliFMM::calibrationState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = HELISTATE_CALIBRATION;
            sEventBroker->post(Event{EV_START_CALIBRATION}, TOPIC_COORDINATION);
            logStatus();
            TRACE("[FMM] Entering calibration state.\n");

            break;
        case EV_EXIT:
            break;
    }
}

void HeliFMM::readyState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            sEventBroker->post(Event{EV_START_SENSOR_LOGGING},
                               TOPIC_COORDINATION);
            status.state = HELISTATE_READY;
            logStatus();
            TRACE("[FMM] Entering ready state.\n");
            break;
        case EV_HELIPIN_DISCONNECTED:
            transition(&HeliFMM::freeFallingState);
            break;
        case EV_EXIT:
            break;
    }
}

void HeliFMM::freeFallingState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state          = HELISTATE_FREEFALL;
            status.time_free_fall = miosix::getTick();
            logStatus();

            delayed_ev_id = sEventBroker->postDelayed(
                {EV_START_CUTTING}, TOPIC_FSM, FREEFALL_TIME_MS);

            TRACE("[FMM] Entering freefall state.\n");
            break;
        case EV_START_CUTTING:
            transition(&HeliFMM::cuttingState);
            break;
        case EV_EXIT:
            break;
    }
}

void HeliFMM::cuttingState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            delayed_ev_id = sEventBroker->postDelayed(
                {EV_TIMEOUT_STOP_CUTTING}, TOPIC_FSM, TIMEOUT_STOP_CUTTING);
            cutter.startCut();

            status.time_start_cut = miosix::getTick();
            status.state          = HELISTATE_CUTTING;
            logStatus();
            TRACE("[FMM] Entering cutting state.\n");

            break;
        case EV_TIMEOUT_STOP_CUTTING:
        case EV_PARACHUTEPIN_DISCONNECTED:
            transition(&HeliFMM::descendingState);
            break;
        case EV_EXIT:
            cutter.stopCut();

            sEventBroker->removeDelayed(delayed_ev_id);
            break;
    }
}

void HeliFMM::descendingState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.time_end_cut = miosix::getTick();
            status.state        = HELISTATE_DESCENDING;
            logStatus();
            TRACE("[FMM] Entering descending state.\n");

            break;
        case EV_BUTTON_LONG_LONG_PRESS:
            transition(&HeliFMM::stoppingState);
            break;
        case EV_EXIT:
            break;
    }
}

void HeliFMM::stoppingState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = HELISTATE_STOPPING;
            logStatus();
#ifdef SD_LOGGING
            log.stop();
#endif
            TRACE("[FMM] Entering stopping state.\n");
            transition(&HeliFMM::stoppedState);
            break;
        case EV_EXIT:
            break;
    }
}

void HeliFMM::stoppedState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = HELISTATE_STOPPED;
            TRACE("[FMM] Entering stopped state.\n");
            break;
        case EV_EXIT:
            break;
    }
}

void HeliFMM::handleEvent(const Event& ev)
{
    // Superclass method
    FSM<HeliFMM>::handleEvent(ev);

    // Update thread stats
    sUpdateDiagnostics(
        StackData{miosix::getTick(), THREAD_FMM,
                  miosix::MemoryProfiling::getCurrentFreeStack(),
                  miosix::MemoryProfiling::getAbsoluteFreeStack()});
}

}  // namespace RogallinaBoard