/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "ScreenManager.h"

#include "debug.h"

#include <miosix.h>
#include <stdio.h>
#include <sstream>
#include <string>

using namespace miosix;
using namespace mxgui;

using miosix::Thread;
using std::string;

const int MARGIN     = 5;
const int INTERSPACE = 13;

static char buffer[256];

// 23 Available text lines
// Screen is 240x320 pixels
namespace RogallinaBoard
{

ScreenManager::ScreenManager(SensorManager* sm, HeliFMM* heli_fmm,
                             PinObserverWrapper* pinobs)
    : display(DisplayManager::instance().getDisplay()), dc(display), dw(dc),
      sm(sm), heli_fmm(heli_fmm), pin_obs(pinobs)
{
    D(printf("Screenmanager constructor\n"));
    clearScreen();
}

ScreenManager::~ScreenManager() {}

void ScreenManager::update()
{
    fetchData();
    // clearScreen();
    drawScreen();
}
void ScreenManager::fetchData()
{
    {
        FastInterruptDisableLock iLock;

        status_logger  = logger.getLogStats();
        logger_started = logger.isStarted();
        logger_file    = logger.getFileName();

        sensorData      = sm->getSensorData();
        logging_sensors = sm->loggingSensors();

        status_fmm = heli_fmm->getStatus();

        pin_status = pin_obs->getPinStatus();
    }

    cpu_usage = sSystemDiagnostics->getCpuUsage();
    min_stack = sSystemDiagnostics->getWorstStack();
}

void ScreenManager::clearScreen()
{
    Color bg = white;
    dw.clearScreen(bg);
}

void ScreenManager::drawScreen()
{
    // Disegna titolo
    writeLine(0, 0, "HELI DROP TEST CONTROLLER", green, black);

    sprintf(buffer, "T: %-10.3f", miosix::getTick() / 1000.0f);

    writeLine(0, 140, buffer);
    drawRogallinaFMM();
    drawLoggerStatus();
    drawSystemStats();
    drawSensorSamples();
}

void ScreenManager::drawRogallinaFMM()
{
    string state;
    Color bgcolor = white, fgcolor = black;

    switch (status_fmm.state)
    {
        case HELISTATE_BOOTED:
            state = "DISARMED";
            break;
        case HELISTATE_CALIBRATION:
            state = "CALIBRATION";
            break;
        case HELISTATE_FAILED:
        {
            bgcolor = red;
            fgcolor = white;
            switch (status_fmm.failReason)
            {
                case FMMFAIL_LOGGER:
                    state = "LOGGER ERROR";
                    break;
                case FMMFAIL_SENSOR:
                    state = "SENSOR INIT ERROR";
                    break;
                default:
                    state = "UNKNOWN ERROR";
                    break;
            }
            break;
        }
        case HELISTATE_READY:
            state = "ARMED";
            break;
        case HELISTATE_FREEFALL:
            state = "State: FREE FALL";
            break;
        case HELISTATE_CUTTING:
            state = "State: CUTTING";
            break;
        case HELISTATE_DESCENDING:
            state = "State: DESCENDING";
            break;
        case HELISTATE_STOPPING:
            state = "State: STOPPING";
            break;
        case HELISTATE_STOPPED:
            bgcolor = blue;
            fgcolor = white;
            state   = "State: STOPPED";
            break;
    }
    
    writeLine(1, 0, "FSM");

    sprintf(buffer, "F: %6d Hz D: %6.3f%% T: %6.3fs", CUTTER_PWM_FREQ, CUTTER_PWM_DUTY*100.0f, TIMEOUT_STOP_CUTTING/1000.0f);
    writeLine(1, 50, buffer);

    sprintf(buffer, "State: %-20s", state.c_str());
    writeLine(2, 0, buffer, fgcolor, bgcolor);

    string pin_state;
    if (pin_status.state == 0)
    {
        bgcolor   = green;
        fgcolor = black;
        pin_state = "CONNECTED   ";
    }
    else
    {
        bgcolor   = red;
        fgcolor = white;
        pin_state = "DISCONNECTED";
    }

    float last_change =
        -(pin_status.last_state_change - miosix::getTick()) / 1000.0f;
    sprintf(buffer, "PIN1: %s", pin_state.c_str());
    writeLine(21, 0, buffer, fgcolor, bgcolor);
    sprintf(buffer, "Num: %3d   (%-5.1fs ago)", pin_status.num_state_changes,
            last_change);
    writeLine(21, 115, buffer);
}

void ScreenManager::drawLoggerStatus()
{
    const int sl = 3;
    const int x1 = 0;
    const int x2 = 120;

    writeLine(sl, x1, "LOGGER");

    sprintf(buffer, "File: %-10s", logger_file.c_str());
    writeLine(sl, x2, buffer);

    int logger_state = 0;  // 0: ready, -1: disabled, 1: started

#ifndef SD_LOGGING
    logger_state = -1;
#else
    if (status_fmm.state == HELISTATE_STOPPED)
        logger_state = 2;
    else
        logger_state = logger_started ? 1 : 0;
#endif
    string state;
    Color fgColor = black;
    Color bgColor = white;
    switch (logger_state)
    {
        case 0:
            state = "READY";
            break;
        case 1:
            bgColor = green;
            if (logging_sensors)
            {
                state = "LOGGING (sens on)";
            }
            else
            {
                state = "LOGGING (sens off)";
            }
            break;
        case 2:
            bgColor = blue;
            state   = "STOPPED";
            break;
        case -1:
            bgColor = red;
            state   = "DISABLED";
            break;
        default:
            bgColor = red;
            state   = "UNKNOWN";
            break;
    }
    sprintf(buffer, "Status: %-20s", state.c_str());
    writeLine(sl + 1, x1, buffer, fgColor, bgColor);

    sprintf(buffer, "Dropped: %8d", status_logger.statDroppedSamples);
    writeLine(sl + 2, x1, buffer);

    sprintf(buffer, "Last error: %8d", status_logger.statWriteError);
    writeLine(sl + 2, x2, buffer);

    sprintf(buffer, "Buf written: %8d", status_logger.statBufferWritten);
    writeLine(sl + 3, x1, buffer);
    sprintf(buffer, "Max time: %8d", status_logger.statMaxWriteTime);
    writeLine(sl + 3, x2, buffer);
}

void ScreenManager::drawSensorSamples()
{
    // Starting line
    const int sl = 6;
    const int x1 = 0;
    const int x2 = 80;
    const int x3 = 160;

    writeLine(sl + 1, 0, "IMU");
    sprintf(buffer, "Compass:  %-8.2f", sensorData.mpu9255_data.compass);
    writeLine(sl + 1, (x2 + x3) / 2, buffer);

    writeLine(sl + 2, x1, "ACC");
    writeLine(sl + 2, x2, "GYRO");
    writeLine(sl + 2, x3, "MAG");

    // X
    sprintf(buffer, "X: %-8.2f", sensorData.mpu9255_data.accel.getX());
    writeLine(sl + 3, x1, buffer);
    sprintf(buffer, "X: %-8.2f", sensorData.mpu9255_data.gyro.getX());
    writeLine(sl + 3, x2, buffer);
    sprintf(buffer, "X: %-8.2f", sensorData.mpu9255_data.magnet.getX());
    writeLine(sl + 3, x3, buffer);

    // Y
    sprintf(buffer, "Y: %-8.2f", sensorData.mpu9255_data.accel.getY());
    writeLine(sl + 4, x1, buffer);
    sprintf(buffer, "Y: %-8.2f", sensorData.mpu9255_data.gyro.getY());
    writeLine(sl + 4, x2, buffer);
    sprintf(buffer, "Y: %-8.2f", sensorData.mpu9255_data.magnet.getY());
    writeLine(sl + 4, x3, buffer);

    // Z
    sprintf(buffer, "Z: %-8.2f", sensorData.mpu9255_data.accel.getZ());
    writeLine(sl + 5, x1, buffer);
    sprintf(buffer, "Z: %-8.2f", sensorData.mpu9255_data.gyro.getZ());
    writeLine(sl + 5, x2, buffer);
    sprintf(buffer, "Z: %-8.2f", sensorData.mpu9255_data.magnet.getZ());
    writeLine(sl + 5, x3, buffer);

    // YAW, PITCH, ROLL
    sprintf(buffer, "YAW: %-8.2f", sensorData.mpu9255_data.yaw);
    writeLine(sl + 6, x1, buffer);
    sprintf(buffer, "PTCH: %-8.2f", sensorData.mpu9255_data.pitch);
    writeLine(sl + 6, x2, buffer);
    sprintf(buffer, "ROLL: %-8.2f", sensorData.mpu9255_data.roll);
    writeLine(sl + 6, x3, buffer);

    // GPS
    writeLine(sl + 7, 0, "GPS");
    sprintf(buffer, "BUF COUNT: %-3d", (int)sensorData.gps_data.buf_count);
    writeLine(sl + 7, x2, buffer);
    // INFO
    switch (sensorData.gps_data.fix)
    {
        case 0:
            sprintf(buffer, "FIX: NOPE    ");
            break;
        case 1:
            sprintf(buffer, "FIX: GPS FIX ");
            break;
        case 2:
            sprintf(buffer, "FIX: AGPS FIX");
            break;
    }
    writeLine(sl + 8, x1, buffer);
    sprintf(buffer, "SATS: %-3d", sensorData.gps_data.sats);
    writeLine(sl + 9, x1, buffer);
    sprintf(buffer, "TIME: %2d:%2d:%2d", sensorData.gps_data.time.hours,
            sensorData.gps_data.time.minutes, sensorData.gps_data.time.seconds);
    writeLine(sl + 10, x1, buffer);

    // COORDS
    sprintf(buffer, "LAT: %-8.2f", sensorData.gps_data.lat);
    writeLine(sl + 8, x2, buffer);
    sprintf(buffer, "LON: %-8.2f", sensorData.gps_data.lon);
    writeLine(sl + 9, x2, buffer);
    sprintf(buffer, "ALT: %-8.2f", sensorData.gps_data.alt);
    writeLine(sl + 10, x2, buffer);

    // DIR, SPEED
    sprintf(buffer, "SPD:  %-8.2f", sensorData.gps_data.speed);
    writeLine(sl + 8, x3, buffer);
    sprintf(buffer, "DIR:  %-8.2f", sensorData.gps_data.bearing);
    writeLine(sl + 9, x3, buffer);
    sprintf(buffer, "L_U:  %-8.1f", sensorData.gps_data.timestamp / 1000.0f);
    writeLine(sl + 10, x3, buffer);

    // BAROMETER
    writeLine(sl + 11, 0, "BAROMETER");

    sprintf(buffer, "PRES:  %-8.2f", sensorData.bme280_data.pressure);
    writeLine(sl + 12, x1, buffer);
    sprintf(buffer, "TEMP:  %-8.2f", sensorData.bme280_data.temperature);
    writeLine(sl + 12, (x2 + x3) / 2, buffer);

    // Current sensor
    sprintf(buffer, "CSENSE: %-8.3f Max: %-8.3f",
            sensorData.current_data.current_value,
            sensorData.current_data.current_max_value);
    writeLine(sl + 13, 0, buffer);
}

void ScreenManager::drawSystemStats()
{
#ifdef DEBUG
    writeLine(22, 0, "DEBUG BUILD. DO NOT FLY.", white, red);
#endif
    sprintf(buffer, "CpuUsage: %4.1f    Worst stack:%2d   %6d bytes", cpu_usage,
            min_stack.thread_id, min_stack.abs_free_stack);
    writeLine(23, 0, buffer);
}

void ScreenManager::write(int x, int y, const char* text, Color txtcolor,
                          Color bgcolor)
{
    dc.setTextColor(txtcolor, bgcolor);
    dc.write(Point(x + MARGIN, y + MARGIN), text);
}
void ScreenManager::writeLine(int line, int x, const char* text, Color txtcolor,
                              Color bgcolor)
{
    write(x, line * INTERSPACE, text, txtcolor, bgcolor);
}
}  // namespace RogallinaBoard
