/*
 * Copyright (c) 2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#pragma once

#include <miosix.h>

#include "HeliTest/config.h"
#include "drivers/pwm/pwm.h"
#include "CutterData.h"
#include "logger/Logger.h"

using miosix::GpioPin;
using miosix::Thread;

namespace RogallinaBoard
{
    
class Cutter
{
public:
    Cutter()
        : pwm(CUTTER_TIM, CUTTER_PWM_FREQ, CUTTER_PWM_DUTY_RESOLUTION),
          pin_enable(CUTTER_ENA::getPin())
    {
        pin_enable.low();

        // Start PWM with 0 duty cycle to keep IN pins low
        pwm.enableChannel(CUTTER_PWM_CHANNEL, 0.0f);

        pwm.start();
    }

    ~Cutter()
    {
        stopCut();
        pwm.stop();
    }

    void startCut()
    {
        if (status.state == CutterState::IDLE)
        {
            enableCutter(CUTTER_PWM_CHANNEL, pin_enable);

            status.state = CutterState::CUTTING;
            logStatus();
        }
    }

    void stopCut()
    {
        if (status.state == CutterState::CUTTING)
        {
            disableCutter(CUTTER_PWM_CHANNEL, pin_enable);

            status.state = CutterState::IDLE;
            logStatus();
        }
    }
private:
    void enableCutter(PWMChannel channel, miosix::GpioPin& ena_pin)
    {
        // Enable PWM Generation
        pwm.setDutyCycle(channel, CUTTER_PWM_DUTY);

        // enable hbridge
        ena_pin.high();
    }

    void disableCutter(PWMChannel channel, miosix::GpioPin& ena_pin)
    {
        pwm.setDutyCycle(channel, 0.0f);  // Set duty cycle to 0 to leave the IN
                                          // pin of the Half-H bridge low

        Thread::sleep(BRIDGE_DISCHARGE_TIME_MS);  // Wait a short delay

        ena_pin.low();  // Disable hbridge
    }

    void logStatus()
    {
        status.timestamp = miosix::getTick();
        Logger::instance().log(status);
    }

    PWM pwm;
    GpioPin pin_enable;

    CutterStatus status;
};

}