/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "RogallinaFMM.h"

#include <stdio.h>
#include "debug.h"

#include "Rogallina/Events.h"
#include "Rogallina/FlightPrograms/ProgramDoNothing.h"
#include "Rogallina/FlightPrograms/ProgramInterface.h"
#include "Rogallina/Servo/ServoControl.h"
#include "SharedThreadIDs.h"
#include "SystemDiagnostics/SystemDiagnostics.h"
#include "events/EventBroker.h"

#include "Rogallina/config.h"

using miosix::Lock;

RogallinaFMM::RogallinaFMM() : FSM(&RogallinaFMM::bootedState)
{
    sEventBroker->subscribe(this, TOPIC_USER_INPUT);
    sEventBroker->subscribe(this, TOPIC_STATUS);
    sEventBroker->subscribe(this, TOPIC_FSM);
}

RogallinaFMM::~RogallinaFMM() {}

FMMStatus RogallinaFMM::getStatus() { return status; }

void RogallinaFMM::bootedState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
        {
            status.state = FMMSTATE_BOOTED;
            logStatus();
            TRACE("Entering booted state.\n");
            break;
        }
        case EV_EXIT:
            break;
        case EV_LOGGER_ERROR:
            transition(&RogallinaFMM::failedState);
            status.failReason = FMMFAIL_LOGGER;
            break;
        case EV_SENSOR_INIT_ERROR:
            transition(&RogallinaFMM::failedState);
            status.failReason = FMMFAIL_LOGGER;
            break;
        case EV_BUTTON_PRESS:
            transition(&RogallinaFMM::programSelectionState);
            break;
        case EV_BUTTON_LONG_PRESS:
            transition(&RogallinaFMM::testServosState);
            break;
#ifdef DEBUG
        case EV_BUTTON_LONG_LONG_PRESS:
            transition(&RogallinaFMM::calibrationState);
            break;
#endif
    }
}

void RogallinaFMM::failedState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = FMMSTATE_FAILED;

            TRACE("Entering failed state.\n");
            break;
        case EV_EXIT:
            break;
    }
}

void RogallinaFMM::testServosState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = FMMSTATE_TEST_SERVOS;
            logStatus();
            wiggleServos();
            TRACE("Entering test servos state.\n");
            break;
        case EV_EXIT:
            break;
    }
}

void RogallinaFMM::calibrationState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = FMMSTATE_CALIBRATION;
            sEventBroker->post(Event{EV_START_CALIBRATION}, TOPIC_COORDINATION);
            logStatus();
            TRACE("Entering calibration state.\n");

            break;
        case EV_EXIT:
            break;
    }
}
void RogallinaFMM::programSelectionState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = FMMSTATE_PROGRAM_SELECT;
            logStatus();
            TRACE("Entering programSelection state.\n");
            break;
        case EV_EXIT:
            break;
        case EV_BUTTON_PRESS:
            // Cycle between available test programs
            status.programID = programManager->nextProgram();
            logStatus();
            break;
        case EV_BUTTON_LONG_PRESS:
            transition(&RogallinaFMM::programReadyState);
            break;
    }
}

void RogallinaFMM::programReadyState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            TRACE("Entering programReady state.\n");
            status.state = FMMSTATE_PROGRAM_READY;
            logStatus();
            sEventBroker->post(Event{EV_PREPARE_PROGRAM}, TOPIC_COORDINATION);
            sEventBroker->post(Event{EV_START_SENSOR_LOGGING},
                               TOPIC_COORDINATION);
            break;
        case EV_EXIT:
            // Start logging the sensors
            break;
        case EV_BUTTON_LONG_PRESS:
#ifdef USE_MAGNETIC_PIN
            transition(&RogallinaFMM::waitingTriggerState);
#else
            transition(&RogallinaFMM::countdownState);
#endif
            break;
    }
}
void RogallinaFMM::countdownState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state            = FMMSTATE_COUNTDOWN;
            status.programStartTime = miosix::getTick() + COUNTDOWN_LENGTH;
            logStatus();

            delayed_ev_id = sEventBroker->postDelayed(
                Event{EV_COUNTDOWN_EXPIRED}, TOPIC_FSM, COUNTDOWN_LENGTH);
            TRACE("Entering countdown state.\n");
            break;
        case EV_EXIT:
            sEventBroker->removeDelayed(delayed_ev_id);
            break;
        case EV_COUNTDOWN_EXPIRED:
            transition(&RogallinaFMM::programExecState);
            break;
    }
}
void RogallinaFMM::waitingTriggerState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = FMMSTATE_WAITING_TRIGGER;
            logStatus();
            TRACE("Entering waitingTrigger state.\n");
            break;
        case EV_EXIT:
            break;
        case EV_MAGPIN_DISCONNECTED:
            transition(&RogallinaFMM::programExecState);
            break;
    }
}
void RogallinaFMM::programExecState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = FMMSTATE_PROGRAM_EXEC;
            logStatus();
            sEventBroker->post(Event{EV_START_PROGRAM}, TOPIC_COORDINATION);
            break;
        case EV_EXIT:
            break;
        case EV_PROGRAM_ENDED:
            transition(&RogallinaFMM::programEndedState);
    }
}
void RogallinaFMM::programEndedState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state            = FMMSTATE_PROGRAM_ENDED;
            status.programEndedTime = miosix::getTick();

            sEventBroker->postDelayed(Event{EV_CLEAN_PROGRAM},
                                      TOPIC_COORDINATION, 15000);
            logStatus();
            TRACE("Entering programEnded state.\n");
            break;
        case EV_EXIT:
            break;
        case EV_BUTTON_LONG_LONG_PRESS:
            transition(&RogallinaFMM::stoppingState);
            break;
    }
}

void RogallinaFMM::stoppingState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = FMMSTATE_STOPPING;
            logStatus();
#ifdef SD_LOGGING
            log.stop();
#endif
            transition(&RogallinaFMM::stoppedState);
            TRACE("Entering stopping state.\n");
            break;
        case EV_EXIT:
            break;
    }
}

void RogallinaFMM::stoppedState(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_ENTRY:
            status.state = FMMSTATE_STOPPED;
            TRACE("Entering stopped state.\n");
            break;
        case EV_EXIT:
            break;
    }
}

void RogallinaFMM::handleEvent(const Event& ev)
{
    // Superclass method
    FSM<RogallinaFMM>::handleEvent(ev);

    // Update thread stats
    sUpdateDiagnostics(
        StackData{miosix::getTick(), THREAD_FMM,
                  miosix::MemoryProfiling::getCurrentFreeStack(),
                  miosix::MemoryProfiling::getAbsoluteFreeStack()});
}

void RogallinaFMM::wiggleServos()
{
    ServoControl* servo = ServoControl::getInstance();
    float wiggle_speed  = 1.7f;
#ifndef ONE_SERVO_AT_A_TIME
    servo->enable();
    servo->start();

    servo->setPositionLeft((float)M_PI, wiggle_speed);
    servo->setPositionRight((float)M_PI, wiggle_speed);
    Thread::sleep(2000);
    servo->setPositionLeft(0, wiggle_speed);
    servo->setPositionRight(0, wiggle_speed);

    servo->stop();
    servo->disable();
#else
    servo->enableLeft(true);
    servo->setPositionLeft((float)M_PI, wiggle_speed);
    Thread::sleep(200);
    servo->setPositionLeft(0, wiggle_speed);

    servo->disableLeft();
    Thread::sleep(500);
    servo->enableRight(true);

    servo->setPositionRight((float)M_PI, wiggle_speed);
    Thread::sleep(200);
    servo->setPositionRight(0, wiggle_speed);
    servo->disableRight();
#endif
}
