/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef SRC_SHARED_BOARDS_ROGALLINA_FLIGHTPROGRAMS_PROGRAMMANAGER_H
#define SRC_SHARED_BOARDS_ROGALLINA_FLIGHTPROGRAMS_PROGRAMMANAGER_H

#include <vector>
#include "ProgramInterface.h"
#include "Singleton.h"
#include "Rogallina/Events.h"
#include "debug.h"
#include "events/FSM.h"

using std::vector;

class ProgramManager : public Singleton<ProgramManager>, public EventHandler
{
    friend class Singleton<ProgramManager>;

public:
    void addProgram(ProgramInterface* program) { programs.push_back(program); }

    size_t nextProgram()
    {
        size_t size = programs.size();
        if (size > 0)
        {
            selectedProgram = (selectedProgram + 1) % size;
        }
        return selectedProgram;
    }

    size_t programsCount() { return programs.size(); }

    size_t selectedProgramIndex() { return selectedProgram; }

    ProgramInterface* getSelectedProgram()
    {
        return programs.at(selectedProgram);
    }

    void execSelectedProgram()
    {
        if (programs.size() > 0)
            programs.at(selectedProgram)->start();
        else
            TRACE("No programs added.\n");
    }

protected:
    void handleEvent(const Event& ev) override
    {
        switch (ev.sig)
        {
            case EV_BUTTON_LONG_PRESS:
                if (getSelectedProgram()->isStarted())
                {
                    getSelectedProgram()->stop();
                }
                break;
            case EV_START_PROGRAM:
                if (!getSelectedProgram()->isStarted())
                {
                    getSelectedProgram()->start();
                }
                break;
            case EV_PREPARE_PROGRAM:
                if (!getSelectedProgram()->isStarted())
                {
                    getSelectedProgram()->prepare();
                }
                break;
            case EV_CLEAN_PROGRAM:
                if (getSelectedProgram()->isStopped())
                {
                    getSelectedProgram()->cleanUp();
                }
                break;
            default:
                break;
        }
    }

private:
    ProgramManager()
    {
        sEventBroker->subscribe(this, TOPIC_COORDINATION);
        sEventBroker->subscribe(this, TOPIC_USER_INPUT);
        sEventBroker->subscribe(this, TOPIC_STATUS);
    }

    vector<ProgramInterface*> programs;
    size_t selectedProgram = 0;
};

#endif /* SRC_SHARED_BOARDS_ROGALLINA_FLIGHTPROGRAMS_PROGRAMMANAGER_H */
