/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "ScreenManager.h"

#include "debug.h"

#include <miosix.h>
#include <stdio.h>
#include <sstream>

using std::stringstream;

using namespace miosix;

using miosix::Thread;
using namespace mxgui;

const int MARGIN     = 5;
const int INTERSPACE = 13;

// 23 Available text lines

// Screen is 240x320 pixels
namespace RogallinaBoard
{

ScreenManager::ScreenManager(SensorManager* sm)
    : display(DisplayManager::instance().getDisplay()), dc(display), dw(dc), sm(sm)
{
    D(printf("Screenmanager constructor\n"));
    clearScreen();
}

ScreenManager::~ScreenManager() {}

void ScreenManager::update()
{
    fetchData();
    // clearScreen();
    drawScreen();
}
void ScreenManager::fetchData()
{
    FastInterruptDisableLock iLock;
    status_logger  = logger.getLogStats();
    logger_started = logger.isStarted();
    logger_file    = logger.getFileName();

    sensorData      = sm->getSensorData();
    logging_sensors = sm->loggingSensors();

    status_fmm = sRogallinaFMM->getStatus();

    cpu_usage = sSystemDiagnostics->getCpuUsage();
    min_stack = sSystemDiagnostics->getWorstStack();
}

void ScreenManager::clearScreen()
{
    Color bg = white;
    dw.clearScreen(bg);
}

void ScreenManager::drawScreen()
{
    // Disegna titolo
    writeLine(0, 0, "ROGALLINA CONTROLLER", green, black);

    char buffer[50];
    sprintf(buffer, "T: %.3f s", miosix::getTick() / 1000.0f);

    writeLine(0, 140, buffer);
    drawRogallinaFMM();
    drawLoggerStatus();
    drawSystemStats();
    drawSensorSamples();
}

void ScreenManager::drawRogallinaFMM()
{
    char buffer[50];
    Color bgcolor = white, fgcolor = black;

    char* progname;
    switch (status_fmm.programID)
    {
        case 0:
            progname = "DN";
            break;
        case 1:
            progname = "TS";
            break;
        case 2:
            progname = "SS";
            break;
        case 3:
            progname = "SSR";
            break;
        case 4:
            progname = "T1";
            break;
    }

    switch (status_fmm.state)
    {
        case FMMSTATE_BOOTED:
            sprintf(buffer, "State: BOOTED");
            break;
        case FMMSTATE_CALIBRATION:
            sprintf(buffer, "State: CALIBRATION");
            break;
        case FMMSTATE_FAILED:
        {
            bgcolor = red;
            fgcolor = white;
            switch (status_fmm.failReason)
            {
                case FMMFAIL_LOGGER:
                    sprintf(buffer, "State: LOGGER ERROR!!!!");
                    break;
                case FMMFAIL_SENSOR:
                    sprintf(buffer, "State: SENSOR INIT ERROR!!!!");
                    break;
                default:
                    sprintf(buffer, "State: UNKNOWN ERROR!!!!");
                    break;
            }
            break;
        }
        case FMMSTATE_PROGRAM_SELECT:
        {
            sprintf(buffer, "State: PROG SELECT    PROG: %s    ", progname);
            break;
        }
        case FMMSTATE_TEST_SERVOS:
        {
            sprintf(buffer, "State: TEST SERVOS    PROG: %s    ", progname);
            break;
        }
        case FMMSTATE_PROGRAM_READY:
        {
            sprintf(buffer, "State: READY           PROG: %s          ",
                    progname);
            break;
        }
        case FMMSTATE_COUNTDOWN:
        {
            float tminus =
                (status_fmm.programStartTime - miosix::getTick()) / 1000.0f;
            sprintf(buffer, "State: COUNTING DOWN PROG %s       T- %.3fs",
                    progname, tminus);
            break;
        }
        case FMMSTATE_WAITING_TRIGGER:
        {

            sprintf(buffer, "State: AWAITING TRIGGER");
            break;
        }
        case FMMSTATE_PROGRAM_EXEC:
            sprintf(buffer, "State: EXECUTING PROG %s                         ",
                    progname);
            break;

        case FMMSTATE_PROGRAM_ENDED:
            sprintf(buffer, "State: PROG ENDED   T=%.3f s      ",
                    status_fmm.programEndedTime / 1000.0f);
            break;
        case FMMSTATE_STOPPING:
            sprintf(buffer, "State: STOPPING     ");
            break;
        case FMMSTATE_STOPPED:
            bgcolor = blue;
            fgcolor = white;
            sprintf(buffer, "State: STOPPED     ");
            break;
    }
    writeLine(1, 0, "STATE MACHINE");
    writeLine(2, 0, buffer, fgcolor, bgcolor);
}
void ScreenManager::drawLoggerStatus()
{
    const int sl = 3;
    const int x1 = 0;
    const int x2 = 120;

    char buffer[50];

    writeLine(sl, x1, "LOGGER");

    sprintf(buffer, "File: %s", logger_file.c_str());
    writeLine(sl, x2, buffer);

    int logger_state = 0;  // 0: ready, -1: disabled, 1: started

#ifndef SD_LOGGING
    logger_state = -1;
#else
    if (status_fmm.state == FMMSTATE_STOPPED)
        logger_state = 2;
    else
        logger_state = logger_started ? 1 : 0;
#endif

    switch (logger_state)
    {
        case 0:
            writeLine(sl + 1, x1, "Status: READY    ");
            break;
        case 1:
            if (logging_sensors)
                writeLine(sl + 1, x1, "Status: LOGGING (sens on) ", black,
                          green);
            else
                writeLine(sl + 1, x1, "Status: LOGGING (sens off) ", black,
                          green);
            break;
        case 2:
            writeLine(sl + 1, x1, "Status: STOPPED        ", white, blue);
            break;
        case -1:
            writeLine(sl + 1, x1, "Status: DISABLED        ", white, red);
            break;
        default:
            writeLine(sl + 1, x1, "Status: UNKNOWN       ", white, red);
            break;
    }

    sprintf(buffer, "Dropped: %d", status_logger.statDroppedSamples);
    writeLine(sl + 2, x1, buffer);

    sprintf(buffer, "Last error: %d", status_logger.statWriteError);
    writeLine(sl + 2, x2, buffer);
}

void ScreenManager::drawSensorSamples()
{
    // Starting line
    const int sl = 6;
    const int x1 = 0;
    const int x2 = 80;
    const int x3 = 160;

    char buffer[30];

    writeLine(sl + 1, 0, "IMU");
    sprintf(buffer, "Compass:  %.2f   ", sensorData.mpu9255_data.compass);
    writeLine(sl + 1, (x2 + x3) / 2, buffer);

    writeLine(sl + 2, x1, "ACC");
    writeLine(sl + 2, x2, "GYRO");
    writeLine(sl + 2, x3, "MAG");

    // X
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.accel.getX());
    writeLine(sl + 3, x1, buffer);
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.gyro.getX());
    writeLine(sl + 3, x2, buffer);
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.magnet.getX());
    writeLine(sl + 3, x3, buffer);

    // Y
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.accel.getY());
    writeLine(sl + 4, x1, buffer);
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.gyro.getY());
    writeLine(sl + 4, x2, buffer);
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.magnet.getY());
    writeLine(sl + 4, x3, buffer);

    // Z
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.accel.getZ());
    writeLine(sl + 5, x1, buffer);
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.gyro.getZ());
    writeLine(sl + 5, x2, buffer);
    sprintf(buffer, "X:  %.2f   ", sensorData.mpu9255_data.magnet.getZ());
    writeLine(sl + 5, x3, buffer);

    // YAW, PITCH, ROLL
    sprintf(buffer, "YAW:  %.2f   ", sensorData.mpu9255_data.yaw);
    writeLine(sl + 6, x1, buffer);
    sprintf(buffer, "PTCH:  %.2f   ", sensorData.mpu9255_data.pitch);
    writeLine(sl + 6, x2, buffer);
    sprintf(buffer, "ROLL:  %.2f   ", sensorData.mpu9255_data.roll);
    writeLine(sl + 6, x3, buffer);

    // GPS
    writeLine(sl + 7, 0, "GPS");
    sprintf(buffer, "BUF COUNT:  %d   ", (int)sensorData.gps_data.buf_count);
    writeLine(sl + 7, x2, buffer);
    // INFO
    switch (sensorData.gps_data.fix)
    {
        case 0:
            sprintf(buffer, "FIX: NOPE");
            break;
        case 1:
            sprintf(buffer, "FIX: GPS FIX");
            break;
        case 2:
            sprintf(buffer, "FIX: AGPS FIX");
            break;
    }
    writeLine(sl + 8, x1, buffer);
    sprintf(buffer, "SATS:  %d   ", sensorData.gps_data.sats);
    writeLine(sl + 9, x1, buffer);
    sprintf(buffer, "TIME:  %2d:%2d:%2d   ", sensorData.gps_data.time.hours,
            sensorData.gps_data.time.minutes, sensorData.gps_data.time.seconds);
    writeLine(sl + 10, x1, buffer);

    // COORDS
    sprintf(buffer, "LAT:  %.2f   ", sensorData.gps_data.lat);
    writeLine(sl + 8, x2, buffer);
    sprintf(buffer, "LON:  %.2f   ", sensorData.gps_data.lon);
    writeLine(sl + 9, x2, buffer);
    sprintf(buffer, "ALT:  %.2f   ", sensorData.gps_data.alt);
    writeLine(sl + 10, x2, buffer);

    // DIR, SPEED
    sprintf(buffer, "SPD:  %.2f   ", sensorData.gps_data.speed);
    writeLine(sl + 8, x3, buffer);
    sprintf(buffer, "DIR:  %.2f   ", sensorData.gps_data.bearing);
    writeLine(sl + 9, x3, buffer);
    sprintf(buffer, "L_U:  %.1f s", sensorData.gps_data.timestamp / 1000.0f);
    writeLine(sl + 10, x3, buffer);

    // BAROMETER
    writeLine(sl + 11, 0, "BAROMETER");

    sprintf(buffer, "PRES:  %.2f   ", sensorData.bme280_data.pressure);
    writeLine(sl + 12, x1, buffer);
    sprintf(buffer, "TEMP:  %.2f   ", sensorData.bme280_data.temperature);
    writeLine(sl + 12, x2, buffer);
}

void ScreenManager::drawSystemStats()
{
    char buffer[70];
    sprintf(buffer, "CpuUsage: %3.1f       Worst stack:%d %d bytes    ",
            cpu_usage, min_stack.thread_id, min_stack.abs_free_stack);
    writeLine(23, 0, buffer);
}

void ScreenManager::write(int x, int y, const char* text, Color txtcolor,
                          Color bgcolor)
{
    dc.setTextColor(txtcolor, bgcolor);
    dc.write(Point(x + MARGIN, y + MARGIN), text);
}
void ScreenManager::writeLine(int line, int x, const char* text, Color txtcolor,
                              Color bgcolor)
{
    write(x, line * INTERSPACE, text, txtcolor, bgcolor);
}
}  // namespace RogallinaBoard
