/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef SRC_SERVOCONTROL_H
#define SRC_SERVOCONTROL_H

#define MAX(x, y) (((x) > (y)) ? (x) : (y))

#include <math.h>
#include <miosix.h>
#include "ServoData.h"
#include "Singleton.h"
#include "Rogallina/config.h"
#include "debug.h"
#include "drivers/servo_stm32.h"
#include "logger/Logger.h"

using miosix::SynchronizedServo;
using miosix::FastMutex;
using miosix::Lock;

static const uint8_t LEFT_SERVO_CHANNEL  = 1;
static const uint8_t RIGHT_SERVO_CHANNEL = 2;

static const float SERVO_TIME_DELTA = 100 / 1000.0f;  // Minimum time delta when
                                                      // updating servo position

static const float MIN_PULSE_WIDTH = 540, MAX_PULSE_WIDTH = 2460;

class ServoControl : public Singleton<ServoControl>
{
    friend class Singleton<ServoControl>;

public:
    ~ServoControl() {}

    void setPositionLeft(float tpos, float speed)
    {
        setPosition(LEFT_SERVO_CHANNEL, tpos, speed);
    }

    void setPositionRight(float tpos, float speed)
    {
        setPosition(RIGHT_SERVO_CHANNEL, tpos, speed);
    }

    void setPositionLeft(float tpos) { setPosition(LEFT_SERVO_CHANNEL, tpos); }

    void setPositionRight(float tpos)
    {
        setPosition(RIGHT_SERVO_CHANNEL, tpos);
    }

    void setPosition(int channel, float tpos, float speed)
    {
        // Do nothing if wrong channel is specified
        if ((channel != LEFT_SERVO_CHANNEL && channel != RIGHT_SERVO_CHANNEL) ||
            speed <= 0)
        {
            return;
        }

        float* cpos = channel == LEFT_SERVO_CHANNEL ? &pos_left : &pos_right;

        if (tpos > (float)M_PI)
        {
            tpos = (float)M_PI;
        }
        else if (tpos < 00)
        {
            tpos = 0;
        }
        speed = fabs(speed);  // Positive speed

        float delta_pos = tpos - *cpos;
        int sign        = delta_pos >= 0 ? 1 : -1;
        float T         = fabs(delta_pos) / speed;
        int steps       = floor(T / SERVO_TIME_DELTA);

        long long t   = miosix::getTick();
        long long end = t + (int)(T * 1000);

        for (long long dt = SERVO_TIME_DELTA * 1000;
             dt < SERVO_TIME_DELTA * 1000 * steps;
             dt += SERVO_TIME_DELTA * 1000)
        {
            setPositionInteral(cpos, channel,
                               *cpos + sign * speed * SERVO_TIME_DELTA);
            Thread::sleepUntil(t + dt);
        }
        Thread::sleepUntil(end);
        setPositionInteral(cpos, channel, tpos);
    }

    void setPosition(int channel, float tpos)
    {
        if (channel != LEFT_SERVO_CHANNEL && channel != RIGHT_SERVO_CHANNEL)
        {
            return;
        }
        float* cpos = channel == LEFT_SERVO_CHANNEL ? &pos_left : &pos_right;
        if (tpos > (float)M_PI)
        {
            tpos = (float)M_PI;
        }
        else if (tpos < 00)
        {
            tpos = 0;
        }

        setPositionInteral(cpos, channel, tpos);
    }

    void start()
    {
        servo.start();
        started = true;
    }
    void stop()
    {
        servo.stop();
        started = false;
    }

    void enable()
    {
        //#ifndef ONE_SERVO_AT_A_TIME
        if (!enabled)
        {
            servo.enable(LEFT_SERVO_CHANNEL);
            servo.enable(RIGHT_SERVO_CHANNEL);
        }
        //#endif
    }

    void enableLeft(bool start = false)
    {
#ifdef ONE_SERVO_AT_A_TIME
        if (!enabled)
        {
            TRACE("Enabling left. S:%d\n", start);
            servo.enable(LEFT_SERVO_CHANNEL);
            enabled = true;
            if (start)
            {
                started = true;
                servo.start();
            }
        }
#endif
    }

    void enableRight(bool start = false)
    {
#ifdef ONE_SERVO_AT_A_TIME
        if (!enabled)
        {
            TRACE("Enabling right. S:%d\n", start);
            servo.enable(RIGHT_SERVO_CHANNEL);
            enabled = true;
            if (start)
            {
                started = true;
                servo.start();
            }
        }
#endif
    }

    void disable()
    {
        //#ifndef ONE_SERVO_AT_A_TIME
        servo.disable(LEFT_SERVO_CHANNEL);
        servo.disable(RIGHT_SERVO_CHANNEL);
        started = false;
        //#endif
    }

    void disableLeft()
    {
#ifdef ONE_SERVO_AT_A_TIME
        if (started)
            servo.stop();
        started = false;
        enabled = false;
        TRACE("Disabling left.\n");
        servo.disable(LEFT_SERVO_CHANNEL);
#endif
    }
    void disableRight()
    {
#ifdef ONE_SERVO_AT_A_TIME
        if (started)
            servo.stop();
        started = false;
        enabled = false;
        TRACE("Disabling right.\n");
        servo.disable(RIGHT_SERVO_CHANNEL);
#endif
    }

    float getPositionLeft() { return pos_left; }
    float getPositionRight() { return pos_right; }

protected:
private:
    ServoControl()
    {
        servo.setMinPulseWidth(MIN_PULSE_WIDTH);
        servo.setMaxPulseWidth(MAX_PULSE_WIDTH);
    }

    void logServoData()
    {
        ServoData data{miosix::getTick(), pos_left, pos_right};
        Logger::instance().log(data);
    }

    /**
     *
     * @param position_ptr
     * @param channel
     * @param tpos
     */
    void setPositionInteral(float* position_ptr, int channel, float tpos)
    {
        TRACE("Setting position: %.1f\n", tpos);
        *position_ptr = tpos;
        servo.setPosition(channel, tpos / (float)M_PI);
        logServoData();
    }

    SynchronizedServo& servo = SynchronizedServo::instance();

    float pos_left  = 0;
    float pos_right = 0;

    bool enabled = false;

    bool started = false;
};

#endif /* SRC_SERVOCONTROL_H */
