/* Copyright (c) 2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <Common.h>
#include <stdexcept>

#include <events/EventBroker.h>
#include <scheduler/TaskScheduler.h>

#include <drivers/HardwareTimer.h>
#include "HeliTest/config.h"

#include "ButtonHandler.h"
#include "HeliTest/Events.h"
#include "HeliTest/FlightModeManager/HeliFMM.h"
#include "HeliTest/PinObserverWrapper.h"
#include "HeliTest/ScreenManager.h"
#include "HeliTest/ThermalCutter/Cutter.h"
#include "HeliTest/config.h"
#include "PinStatus.h"
#include "SensorManager/SensorManager.h"
#include "SharedThreadIDs.h"
#include "SystemDiagnostics/SystemDiagnostics.h"
#include "hwmapping.h"

using std::runtime_error;

using namespace miosix;
using namespace RogallinaBoard;
using namespace RogallinaBoard::Sensors;
using namespace interfaces;

using ButtonHandler_t = ButtonHandler<interfaces::buttons::user>;

ButtonHandler_t* btnHandler;
TaskScheduler* scheduler;
SensorManager* sensorManager;
ScreenManager* screen;

Cutter* cutter;
HeliFMM* fmm;
PinObserverWrapper* pinObserver;
Logger& logger = Logger::instance();

void pinCallback(unsigned int port, unsigned int num);

void onButtonPressCallback(ButtonHandler_t::ButtonPress type);

void initPins()
{
    {
        FastInterruptDisableLock d;

        DETACHMENT_PIN::mode(DETACHMENT_PIN_MODE);
        CUTTER_ENA::mode(Mode::OUTPUT);

        CUTTER_IN::mode(Mode::ALTERNATE);
        CUTTER_IN::alternateFunction(2);

        CUTTER_CSENSE::mode(Mode::INPUT_ANALOG);
    }

    interfaces::buttons::user::mode(Mode::INPUT);
    red_led::mode(Mode::OUTPUT);
    green_led::mode(Mode::OUTPUT);

    // Debug enabled warning
    D(red_led::high());
}

void initComponents()
{
    EventBroker::getInstance();

    btnHandler    = new ButtonHandler_t(1, &onButtonPressCallback);
    scheduler     = new TaskScheduler(); 
    cutter        = new Cutter();
    fmm           = new HeliFMM(*cutter);
    sensorManager = new SensorManager(scheduler);
    pinObserver   = new PinObserverWrapper();

    screen = new ScreenManager(sensorManager, fmm, pinObserver);
}

int main()
{
    initPins();

    initComponents();

    // Start the active objects
    scheduler->start(4096, 2);
    sEventBroker->start(2048, 3);

#ifdef SD_LOGGING
    try
    {
        logger.start();
    }
    catch (runtime_error& err)
    {
        TRACE("Error opening log!\n");
        sEventBroker->post(Event{EV_LOGGER_ERROR}, TOPIC_STATUS);
    }
#endif

    btnHandler->start(2048, 1);
    fmm->start(4096, 3);
    sensorManager->start(4096, 1);

    pinObserver->start(2048, 3);

    sEventBroker->postDelayed({EV_START_SAMPLING}, TOPIC_COORDINATION, 100);

    while (true)
    {
        // Refresh the screen at 5Hz
        screen->update();

        sSystemDiagnostics->logSystemData();
        sSystemDiagnostics->updateThreadDiagnostics(
            StackData{miosix::getTick(), THREAD_MAIN,
                      MemoryProfiling::getCurrentFreeStack(),
                      MemoryProfiling::getAbsoluteFreeStack()});

        Thread::sleep(200);
    }
}

void onButtonPressCallback(ButtonHandler_t::ButtonPress type)
{

    switch (type)
    {
        case ButtonHandler_t::ButtonPress::SHORT:
            sEventBroker->post({EV_BUTTON_PRESS}, TOPIC_USER_INPUT);
            break;
        case ButtonHandler_t::ButtonPress::LONG:
            sEventBroker->post({EV_BUTTON_LONG_PRESS}, TOPIC_USER_INPUT);
            break;
        case ButtonHandler_t::ButtonPress::VERY_LONG:
            sEventBroker->post({EV_BUTTON_LONG_LONG_PRESS}, TOPIC_USER_INPUT);
            break;

        default:
            break;
    }
}