/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <Common.h>
#include <stdexcept>

#include "ButtonHandler.h"
#include "Rogallina/Events.h"
#include "Rogallina/FlightModeManager/RogallinaFMM.h"
#include "Rogallina/FlightPrograms/FlightPrograms.h"
#include "Rogallina/FlightPrograms/ProgramManager.h"
#include "SharedThreadIDs.h"
#include "Rogallina/ScreenManager.h"
#include "SensorManager/SensorManager.h"
#include "Rogallina/Servo/ServoControl.h"
#include "Rogallina/config.h"
#include "hwmapping.h"
#include "SystemDiagnostics/SystemDiagnostics.h"
#include "events/EventBroker.h"
#include "scheduler/TaskScheduler.h"

using namespace miosix;
using namespace RogallinaBoard;
using namespace RogallinaBoard::Sensors;

using ButtonHandler_t = ButtonHandler<interfaces::buttons::user>;

ButtonHandler_t* btnHandler;
ScreenManager* screen;
SensorManager* sensorManager;
RogallinaFMM* fmm;
ProgramManager* programManager;
TaskScheduler* scheduler;
Logger& logger = Logger::instance();

void onButtonPressCallback(ButtonHandler_t::ButtonPress type)
{

    switch (type)
    {
        case ButtonHandler_t::ButtonPress::SHORT:
            sEventBroker->post({EV_BUTTON_PRESS}, TOPIC_USER_INPUT);
            break;
        case ButtonHandler_t::ButtonPress::LONG:
            sEventBroker->post({EV_BUTTON_LONG_PRESS}, TOPIC_USER_INPUT);
            break;
        case ButtonHandler_t::ButtonPress::VERY_LONG:
            sEventBroker->post({EV_BUTTON_LONG_LONG_PRESS}, TOPIC_USER_INPUT);
            break;

        default:
            break;
    }
}

void initComponents()
{
    EventBroker::getInstance();

    scheduler      = new TaskScheduler();
    btnHandler     = new ButtonHandler_t(1, &onButtonPressCallback);
    sensorManager  = new SensorManager(scheduler);
    screen         = new ScreenManager(sensorManager);
    fmm            = RogallinaFMM::getInstance();
    programManager = ProgramManager::getInstance();

    programManager->addProgram(new ProgramDoNothing());
    programManager->addProgram(new ProgramTestServos());
    programManager->addProgram(new ProgramSuperStall());
    programManager->addProgram(new ProgramSuperStallRecovery());
    programManager->addProgram(new ProgramTest1());
}

int main()
{
    red_led::mode(Mode::OUTPUT);
    green_led::mode(Mode::OUTPUT);

    // Debug enabled warning
    D(red_led::high());

    initComponents();

    // Start the active objects
    scheduler->start();
    sEventBroker->start();

#ifdef SD_LOGGING
    try
    {
        logger.start();
    }
    catch (std::runtime_error& err)
    {
        TRACE("Error opening log!\n");
        sEventBroker->post(Event{EV_LOGGER_ERROR}, TOPIC_STATUS);
    }
#endif

    btnHandler->start();
    fmm->start();
    programManager->start();
    sensorManager->start();

    sEventBroker->postDelayed(Event{EV_START_SAMPLING}, TOPIC_COORDINATION,
                              1000);

    while (true)
    {
        // Refresh the screen at 5Hz
        screen->update();

        sSystemDiagnostics->logSystemData();
        sSystemDiagnostics->updateThreadDiagnostics(
            StackData{miosix::getTick(), THREAD_MAIN,
                      MemoryProfiling::getCurrentFreeStack(),
                      MemoryProfiling::getAbsoluteFreeStack()});

        Thread::sleep(200);
    }
}
