/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#pragma once

#include <functional>

#include "ActiveObject.h"
#include "ButtonStatus.h"
#include "logger/Logger.h"

using miosix::Thread;

static const int TICK_LENGTH           = 100;  // milliseconds
static const int LONG_PRESS_TICKS      = 10;
static const int VERY_LONG_PRESS_TICKS = 50;

/**
 * @brief Class that detects if a button is pressed, long pressed or long-long
 * pressed and calls a callback in each case
 *
 * @tparam Button GPIO of the button
 */
template <typename Button>
class ButtonHandler : public ActiveObject
{
public:
    enum class ButtonPress
    {
        SHORT,
        LONG,
        VERY_LONG
    };

    using ButtonCallback = std::function<void(ButtonPress)>;

    static bool isPressed() { return Button::value(); }

    ButtonHandler(uint8_t btn_id, ButtonCallback on_press)
        : ButtonHandler(btn_id, on_press, on_press, on_press)
    {
    }

    ButtonHandler(uint8_t btn_id, ButtonCallback on_short_press,
                  ButtonCallback on_long_press,
                  ButtonCallback on_very_long_press)
        : on_short_press(on_short_press), on_long_press(on_long_press),
          on_very_long_press(on_very_long_press)
    {
        memset(&status, 0, sizeof(ButtonStatus));
        status.btn_id = btn_id;
        Button::mode(miosix::Mode::INPUT);
    }
    ~ButtonHandler() {}

protected:
    void run() override
    {
        while (!shouldStop())
        {
            if (Button::value())
            {
                status.pressed_ticks++;
                status.pressed   = true;
                status.timestamp = miosix::getTick();

                Logger::instance().log(status);
            }
            else if (status.pressed)  // if the button was unpressed since the
                                      // last operation
            {
                if (status.pressed_ticks >= VERY_LONG_PRESS_TICKS)
                {
                    TRACE("Button pressed (very long) (%d ticks).\n",
                          status.pressed_ticks);
                    if (on_very_long_press)
                    {
                        on_very_long_press(ButtonPress::VERY_LONG);
                    }
                }
                else if (status.pressed_ticks >= LONG_PRESS_TICKS)
                {
                    TRACE("Button pressed (long) (%d ticks).\n",
                          status.pressed_ticks);
                    if (on_long_press)
                    {
                        on_long_press(ButtonPress::LONG);
                    }
                }
                else
                {
                    TRACE("Button pressed (short) (%d ticks).\n",
                          status.pressed_ticks);
                    if (on_short_press)
                    {
                        on_short_press(ButtonPress::SHORT);
                    }
                }
                status.pressed       = false;
                status.pressed_ticks = 0;
                status.timestamp     = miosix::getTick();
                Logger::instance().log(status);
            }

            Thread::sleep(TICK_LENGTH);
        }
    }

private:
    ButtonCallback on_short_press;
    ButtonCallback on_long_press;
    ButtonCallback on_very_long_press;

    ButtonStatus status;
};