/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "SensorManager.h"

#include <events/EventBroker.h>

#include "SensorManagerConfig.h"
#include "SharedConfig.h"
#include "SharedEvents.h"
#include "SharedThreadIDs.h"
#include "SystemDiagnostics/SystemDiagnostics.h"

namespace RogallinaBoard
{
namespace Sensors
{

using miosix::FastMutex;
using miosix::Lock;

SensorManager::SensorManager(TaskScheduler* scheduler)
    : logger(Logger::instance()), scheduler(scheduler)
{
    sEventBroker->subscribe(this, TOPIC_COORDINATION);

    initSensors();
    initSamplers();

    scheduler->add(
        [] {
            // Update thread stats
            sUpdateDiagnostics(
                StackData{miosix::getTick(), THREAD_SCHEDULER,
                          miosix::MemoryProfiling::getCurrentFreeStack(),
                          miosix::MemoryProfiling::getAbsoluteFreeStack()});
        },
        1000, TASK_THREADS);  // Log diagnostics @ 1Hz
}
SensorManager::~SensorManager()
{
    if (bme280 != nullptr)
        delete bme280;
    if (gps != nullptr)
        delete gps;
    if (mpu9255 != nullptr)
        delete mpu9255;
    if (current_sensor != nullptr)
        delete current_sensor;
}

bool SensorManager::init()
{
    bool success = true;

    success = success && initSensors();
    success = success && initSamplers();

    return success;
}

bool SensorManager::initSensors()
{
    bool success = true;

    bme280 = new BME280Type();
    if (!bme280->init())
    {
        TRACE("Init error: BME280\n");
        success = false;
    }

    mpu9255 = new MPU9255(IMU_SAMPLE_RATE);
    if (!mpu9255->init())
    {
        TRACE("Init error: MPU9255\n");
        success = false;
    }

    gps = new GPS(1);
    if (!gps->init())
    {
        TRACE("Init error: GPS\n");
        success = false;
    }

    current_sensor = new CurrentSensor();
    if (!current_sensor->init())
    {
        TRACE("Init error: CURRENT SENSE\n");
        success = false;
    }

    return success;
}

bool SensorManager::initSamplers()
{
    sampler_press_simple.AddSensor(bme280);
    sampler_imu_simple.AddSensor(mpu9255);
    sampler_gps_simple.AddSensor(gps);
    sampler_current_simple.AddSensor(current_sensor);
    return true;
}

void SensorManager::handleEvent(const Event& ev)
{
    switch (ev.sig)
    {
        case EV_START_SAMPLING:
            startSampling();
            break;
        case EV_START_SENSOR_LOGGING:
            logSensors = true;
            break;
        case EV_START_CALIBRATION:
            TRACE("Enabling calibrations prints...\n");
            mpu9255->resetMagnetometerCalibration();
            printCalibration = true;
            break;
        default:
            TRACE("Unrecognized event\n");
    }

    // Update thread stats
    sUpdateDiagnostics(
        StackData{miosix::getTick(), THREAD_SENSOR_MANAGER,
                  miosix::MemoryProfiling::getCurrentFreeStack(),
                  miosix::MemoryProfiling::getAbsoluteFreeStack()});
}

SensorData SensorManager::getSensorData() { return data; }

void SensorManager::startSampling()
{
    /*
     * std::bind syntax:
     * std::bind(&MyClass::someFunction, &myclass_instance, [someFunction args])
     */

    // Simple Pressure Sampler callback and scheduler function
    std::function<void()> simple_press_callback =
        std::bind(&SensorManager::onSimplePressCallback, this);

    std::function<void()> simple_press_sampler =
        std::bind(&SimpleSensorSampler::UpdateAndCallback,
                  &sampler_press_simple, simple_press_callback);

    scheduler->add(simple_press_sampler, 1000 / PRESS_SAMPLE_RATE,
                   TASK_PRESSURE);

    // Simple Current Sampler callback and scheduler function
    std::function<void()> simple_current_callback =
        std::bind(&SensorManager::onSimpleCurrentCallback, this);

    std::function<void()> simple_current_sampler =
        std::bind(&SimpleSensorSampler::UpdateAndCallback,
                  &sampler_current_simple, simple_current_callback);

    scheduler->add(simple_current_sampler, 1000 / ADC_SAMPLE_RATE, TASK_ADC);

    // Simple IMU Sampler callback and scheduler function
    std::function<void()> simple_imu_callback =
        std::bind(&SensorManager::onSimpleIMUCallback, this);

    std::function<void()> simple_imu_sampler =
        std::bind(&SimpleSensorSampler::UpdateAndCallback, &sampler_imu_simple,
                  simple_imu_callback);

    scheduler->add(simple_imu_sampler, 1000 / IMU_SAMPLE_RATE, TASK_IMU);

    // Simple GPS Sampler callback and scheduler function

    // Start gps AO
    TRACE("Starting GPS\n");
    gps->start();

    std::function<void()> simple_gps_callback =
        std::bind(&SensorManager::onSimpleGPSCallback, this);

    std::function<void()> simple_gps_sampler =
        std::bind(&SimpleSensorSampler::UpdateAndCallback, &sampler_gps_simple,
                  simple_gps_callback);

    scheduler->add(simple_gps_sampler, 1000 / GPS_SAMPLE_RATE, TASK_GPS);
}

void SensorManager::onSimplePressCallback()
{
    // Log data & update sensor data struct
    data.bme280_data = *bme280->getDataPtr();

    if (logSensors)
        logger.log(*bme280->getDataPtr());
}

void SensorManager::onSimpleIMUCallback()
{
    // Log data & update sensor data struct
    data.mpu9255_data = *mpu9255->getIMUDataPtr();

#ifdef ENABLE_CALIBRATION
    if (printCalibration)
        mpu9255->printDataForCalibration();
#endif

    if (logSensors)
        logger.log(*mpu9255->getIMUDataPtr());
}

void SensorManager::onSimpleCurrentCallback()
{
    // Log data & update sensor data struct
    data.current_data = *(current_sensor->getCurrentDataPtr());

    if (logSensors)
        logger.log(data.current_data);
}

void SensorManager::onSimpleGPSCallback()
{
    // Log data & update sensor data struct
    data.gps_data = *gps->getGPSDataPtr();
    if (logSensors)
        logger.log(*gps->getGPSDataPtr());
}

}  // namespace Sensors
}  // namespace RogallinaBoard
