/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#pragma once

#include <events/FSM.h>
#include <logger/Logger.h>
#include <scheduler/TaskScheduler.h>
#include <sensors/BME280/BME280.h>

#include "Sensors/MPU9255.h"
#include "Sensors/gps.h"
#include "hwmapping.h"
#include "sensors/SensorSampling.h"
#include "Sensors/CurrentSensor.h"

// Forward declarations
typedef BME280<RogallinaBoard::interfaces::sensors::bme280::spi> BME280Type;

// Type definitions
namespace RogallinaBoard
{
namespace Sensors
{

/**
 * Common storage point for all the latest samples of each sensor.
 */
struct SensorData
{
    BME280Data bme280_data;
    MPU9255Data mpu9255_data;
    GPSData gps_data;
    CurrentSenseData current_data;
};

/**
 * The SensorManager class manages all the sensors connected to the board.
 *
 * Sensors are grouped by "type" (Simple or DMA) and "sample frequency" and
 * grouped in various SensorSampler objects. These objects are then added to the
 * scheduler that manages the timings for the sampling.
 * After a SensorSampler has finished sampling its sensors, it will call a
 * callback, where these samples can be processed and dispatched.
 */
class SensorManager : public EventHandler
{
public:
    /**
     * Returns a copy of the struct containing the latest available samples
     * from each sensor at the time of the call.
     * This function is not thread-safe: it must be called in the scheduler
     * thread or with interrupt disabled.
     */
    SensorData getSensorData();

    bool init();

    bool loggingSensors() { return logSensors; }

    SensorManager(TaskScheduler* scheduler);
    ~SensorManager();

private:
    /**
     * Initialize all the sensors.
     */
    bool initSensors();

    /**
     * Initialize the samplers.
     */
    bool initSamplers();

    /**
     * Handle received events.
     * @param ev Event to handle.
     */
    void handleEvent(const Event& ev);

    /**
     * Adds all the SensorSamplers to the scheduler and begins sampling.
     */
    void startSampling();

    /*
     * Callbacks. These functions are called each time the corresponding
     * SensorSampler has acquired new samples.
     * These functions are called on the scheduler (sampler) thread, so avoid
     * performing non-critical, intensive tasks.
     */

    /**
     * Simple, 5 Hz SensorSampler Callback.
     */
    void onSimpleGPSCallback();

    /**
     * Simple, 50 Hz SensorSampler Callback.
     */
    void onSimplePressCallback();

    /**
     * Simple, 20 Hz Current sense Callback.
     */
    void onSimpleCurrentCallback();

    /**
     * Simple, 200 Hz SensorSampler Callback.
     */
    void onSimpleIMUCallback();

    // Sensor samplers
    SimpleSensorSampler sampler_gps_simple;
    SimpleSensorSampler sampler_press_simple;
    SimpleSensorSampler sampler_imu_simple;
    SimpleSensorSampler sampler_current_simple;

    // Sensors
    BME280Type* bme280 = nullptr;
    MPU9255* mpu9255   = nullptr;
    GPS* gps           = nullptr;
    CurrentSensor* current_sensor = nullptr;

    SensorData data;

    bool logSensors       = false;  // Don't log sensor data until required
    bool printCalibration = false;

    // Logger ref
    Logger& logger;

    TaskScheduler* scheduler;
};

}  // namespace Sensors
}  // namespace RogallinaBoard
