#include "MPU9255.h"
#include "debug.h"
/*
 ******************************************************************************
 * CONSTRUCTOR
 ******************************************************************************
 */

MPU9255::MPU9255(float imu_freq)
{
    assetto     = new Mahony();
    sample_freq = imu_freq;
}

/*
 ******************************************************************************
 * INITIALIZING METHODS
 ******************************************************************************
 */

/*
 * La funzione che chiama tutti gli inizializzatori, estende la classe
 * basesensors, quindi può essere chiamata nel loop delle inizializzazioni
 * dei sensori vettorizzati.
 */
bool MPU9255::init()
{
    bool result = true;
    i2cInit();
    Thread::sleep(10);
    
    if (!i2cReadIDTest())
    {
        return false;
    }
    axelInit(AFS_16G);  // inizializzo l'acceleromerto
    tempInit();
    result = result && gyroInit(GFS_2000DPS);  // inizializzo il giroscopio
    configPassThrough();  // Necessario per accedere direttamente al
                          // magnetometro, che è nello stesso chip ma a un
                          // indirizzo differente
    readMagAdj();  // deve essere chiamata sempre PRIMA di configMagMode perchè
                   // reimposta il valore della modialità lettura in ROM_MODE
    configMagMode(MPU_AK89XX_CONT2_MODE);  // imposta la modalità di lettura
                                           // continua a 100hz
    readIDMag();
    assetto->begin(sample_freq);  // inizializzo la funzione che calcola
                                  // l'assetto con la frequenza di campionamento
                                  // impostata
    TRACE("IMU initialized, result = %s\n", result ? "ok" : "error");

    if (result)
        initialized = true;

    return result;
}

void MPU9255::i2cInit()
{
    soft_i2c::init();
    TRACE("\nI2C inizialized...");
}

void MPU9255::axelInit(axelFSRchoice fsr)
{
    axelFSR = fsr;

    uint8_t config = (uint8_t)axelFSR;

    // leggo i dati di configurazione offset l'accelerometro
    masterWrite(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_ACCEL_CONFIG, &config, 1);
    rawAxelOffset[0] =
        static_cast<uint16_t>(axel_x_offset * sensor_fsr / getAxelFSR());
    rawAxelOffset[1] =
        static_cast<uint16_t>(axel_y_offset * sensor_fsr / getAxelFSR());
    rawAxelOffset[2] =
        static_cast<uint16_t>(axel_z_offset * sensor_fsr / getAxelFSR());
    TRACE("\nAXEL OFFSET: X%d, Y%d, Z%d", rawAxelOffset[0], rawAxelOffset[1],
          rawAxelOffset[2]);

    // TODO ricalibrare l'accelerometro e inviare la calibrazione corretta nel
    // registro
    config = 0x01;  // configuro il low pass filter
    masterWrite(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_ACCEL_CONFIG_2, &config, 1);
    TRACE("\nAXEL inizialized to %dG FSR and 184Hz LPF...", getAxelFSR());
}

void MPU9255::tempInit()
{
    temp_sensitivity = 1;
    room_temp_offset = 0;
    TRACE("\nTEMP sensor inizialized...");
}

bool MPU9255::gyroInit(gyroFSRchoice fsr)
{
    gyroFSR = fsr;

    uint8_t config = (uint8_t)gyroFSR;
    masterWrite(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_GYRO_CONFIG, &config, 1);
    config = 0x43;  // configuro il low pass filter a 41Hz e la possibilità di
                    // sovrascrivere i vecchi dati nella FIFO
    masterWrite(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_CONFIG, &config, 1);
    bool result = gyroOffsetReg();
    TRACE("\nGYRO inizialized to %dDPS FSR and 41Hz LPF...", getGyroFSR());
    return result;
}

bool MPU9255::gyroOffsetReg()
{
    /* FS_SEL | Full Scale Range   | LSB Sensitivity
    * -------+--------------------+----------------
    * 0      | +/- 250 degrees/s  | 131 LSB/deg/s
    * 1      | +/- 500 degrees/s  | 65.5 LSB/deg/s
    * 2      | +/- 1000 degrees/s | 32.8 LSB/deg/s
    * 3      | +/- 2000 degrees/s | 16.4 LSB/deg/s
    */

    uint8_t sens_reg[NUM_RAW_UINT8];
    const int16_t *sens_reg_ptr = (const int16_t *)sens_reg;
    int16_t temp_raw_gyro[3];
    uint8_t off_reg[6];
    const uint16_t *off_reg_ptr = (const uint16_t *)off_reg;
    uint16_t i, j = 0;
    float offset_mean[3] = {gyro_x_offset, gyro_y_offset, gyro_z_offset};

    // ricevo gli offset interni del gyro (sono impostabili da utente,
    // quindi hanno importanza discutibile)
    masterRead(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_XG_OFFSET_H, off_reg, 6);

    rawGyroOffset[0] = fromBigEndian16(off_reg_ptr[0]);
    rawGyroOffset[1] = fromBigEndian16(off_reg_ptr[1]);
    rawGyroOffset[2] = fromBigEndian16(off_reg_ptr[2]);

    for (i = 0; i < 100;
         i++)  // faccio n campionamenti per mediare il valore zero
    {
        masterRead(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_ACCEL_XOUT_H, sens_reg,
                   NUM_RAW_UINT8);  // ricevo tutti i dati
        temp_raw_gyro[0] = fromBigEndian16(sens_reg_ptr[4]);
        temp_raw_gyro[1] = fromBigEndian16(sens_reg_ptr[5]);
        temp_raw_gyro[2] = fromBigEndian16(sens_reg_ptr[6]);
        for (j = 0; j < 3; j++)
            offset_mean[j] =
                (offset_mean[j] * 63 + (temp_raw_gyro[j] - rawGyroOffset[j])) /
                64;  // media mobile pesata su 64 valori
        if (!(i % 20))
            TRACE("\nGYRO-OFFSET: X% 3.2f Y% 3.2f Z% 3.2f",
                  offset_mean[0] + rawGyroOffset[0],
                  offset_mean[1] + rawGyroOffset[1],
                  offset_mean[2] + rawGyroOffset[2]);
        Thread::sleep(10);  // distanti 10ms tra di loro
    }
    for (i = 0; i < 3; i++)
    {
        rawGyroOffset[i] += static_cast<int16_t>(
            offset_mean[i]);  // sommo ai valori di calibrazione interna
        TRACE("\nGYRO-OFFSET: %d :%d", i, rawGyroOffset[i]);
        if (rawGyroOffset[i] > 0.05 * getGyroFSR())  // corrisponde a un valore
                                                     // che è sicuramente
                                                     // fuoriscala per qualsiasi
                                                     // tipo di test
        {
            TRACE(
                "\n\n\nWARNING!!!!! CALIBRATION CANNOT CONTINUE: THE GYRO IS "
                "MOVING. PLEASE KEEP IT STILL ON THE GROUND.\n\n\n");
            return false;
            /*while (true)
                Thread::sleep(100);*/
        }
    }
    gyro_cal_failed = false;
    return true;
    // TODO: riscrivere nella memoria del sensore i nuovi valori di calibrazione
}

void MPU9255::configPassThrough()
{
    uint8_t config = MPU_BYPASS_ENABLE;
    masterWrite(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_INT_PIN_CFG, &config, 1);
    // TRACE("\n\nI2C Passthrough set...");
}

void MPU9255::readMagAdj()
{
    configMagMode(MPU_AK89XX_ROM_ACC_MODE);  // per poter leggere i dati
                                             // corretti dalla rom
    masterRead(AK8963_I2C_7BIT_ADDRESS, MPU_AK89XX_REG_ASAX, mag_adj,
               3);  // mag_adj è un vettore da 3 valori
    TRACE("\nMAG ADJ values: X%d Y%d Z%d", mag_adj[0], mag_adj[1], mag_adj[2]);
}

void MPU9255::configMagMode(uint8_t mag_mode)
{
    /*   0001XXXX per modalità a 16 bit, con XXXX:
     *      "0000": Power-down mode
     *      "0001": Single measurement mode
     *      "0010": Continuous measurement mode 1 (8hz)
     *      "0110": Continuous measurement mode 2 (100hz)
     *      "0100": External trigger measurement mode
     *      "1000": Self-test mode
     *      "1111": Fuse ROM access mode
     */
    uint8_t config = MPU_AK89XX_OFF_MODE;
    std::string buffer;
    // per poter cambiare la mode si deve per forza transitare da off mode prima
    masterWrite(AK8963_I2C_7BIT_ADDRESS, MPU_AK89XX_REG_CNTL, &config, 1);
    config = mag_mode;
    // config1 è composto da 1 solo byte
    masterWrite(AK8963_I2C_7BIT_ADDRESS, MPU_AK89XX_REG_CNTL, &config, 1);

    /*switch (mag_mode)
    {
        case MPU_AK89XX_OFF_MODE:
            buffer = "\nMAG in POWER SAVE mode";
            break;
        case MPU_AK89XX_SINGLE_MODE:
            buffer = "\nMAG in SINGLE mode";
            break;
        case MPU_AK89XX_CONT1_MODE:
            buffer = "\nMAG in CONTINUOUS 1 mode";
            break;
        case MPU_AK89XX_EXT_TR_MODE:
            buffer = "\nMAG in EXTERNAL TRIGGER mode";
            break;
        case MPU_AK89XX_CONT2_MODE:
            buffer = "\nMAG in CONTINUOUS 2 mode";
            break;
        case MPU_AK89XX_SELF_TEST_MODE:
            buffer = "\nMAG in SELF TEST mode";
            break;
        case MPU_AK89XX_ROM_ACC_MODE:
            buffer = "\nMAG in FUSE ACCESS mode";
            break;
        default:
            buffer = "\n\nERROR in MAG mode!!!\n\n";
            break;
    }

    TRACE("%s", buffer.c_str());
    fflush(stdout);*/
}

/*
 ******************************************************************************
 * VALUES UPDATING METHODS
 ******************************************************************************
 */

/*
 * Chiamata ogni volta che faccio una lettura del magnetometro, aggiusta i
 * nuovi valori con i valori di calibrazione ottenuti one-time
 */
void MPU9255::magCalibration()
{
    float V[3];

    V[0] = data_out[7] - mag_cal_offset[0];  // /  sensor_fsr   * 4800);
                                             // //modificato da teo per avere i
                                             // valori di calibrazione in uT
    V[1] = data_out[8] - mag_cal_offset[1];  // /  sensor_fsr   * 4800);
                                             // //modificato da teo per avere i
                                             // valori di calibrazione in uT
    V[2] = data_out[9] - mag_cal_offset[2];  // /  sensor_fsr   * 4800);
                                             // //modificato da teo per avere i
                                             // valori di calibrazione in uT

    // Moltiplico per la matrice di calibrazione
    data_out[7] = V[0] * mag_cal_matrix[0][0] + V[1] * mag_cal_matrix[0][1] +
                  V[2] * mag_cal_matrix[0][2];
    data_out[8] = V[0] * mag_cal_matrix[1][0] + V[1] * mag_cal_matrix[1][1] +
                  V[2] * mag_cal_matrix[1][2];
    data_out[9] = V[0] * mag_cal_matrix[2][0] + V[1] * mag_cal_matrix[2][1] +
                  V[2] * mag_cal_matrix[2][2];
}

/*
 *  Chiamata per andare a leggere i nuovi valori dal magnetometro e salvarli nel
 * registro
 */
bool MPU9255::updateMagValues()
{
    uint8_t sensreg[6];
    int16_t *sens_reg_ptr = (int16_t *)sensreg;
    uint8_t status2_reg   = 0;

    masterRead(AK8963_I2C_7BIT_ADDRESS, MPU_AK89XX_REG_HXL, sensreg, 6);

    // questo registro DEVE essere chiamato dopo la lettura,
    // altrimenti non viene "terminata" correttamente
    masterRead(AK8963_I2C_7BIT_ADDRESS, MPU_AK89XX_REG_ST2, &status2_reg, 1);

    if (status2_reg & 0x08)  // se c'è stato un overflow magnetico
    {
        // TRACE("\n\nERROR: MAGNETIC OVERFLOW\n\n");
        return false;  // i dati appena letti non mi interessano ed esco
    }

    /*ATTENZIONE: IL MAGNETOMETRO NON E' BIG ENDIAN COME
     * L'ACCELEROMETRO/GIROSCOPIO!!!!!*/
    // MAG (sensreg è da 8 bit(((int16_t)sensreg[0]) |
    // ((int16_t)sensreg[1]<<8)), li unisco in una sola da 16)
    rawdata[7] = fromLittleEndian16(sens_reg_ptr[1]);  // ASSE X, è l'asse Y del
                                                       // magnetometro, per fare
                                                       // in modo di allineare
                                                       // la terna con quella
                                                       // del resto del chip
    rawdata[8] = fromLittleEndian16(
        sens_reg_ptr[0]);  // ASSE Y, è l'asse X del magnetometro
    rawdata[9] =
        -1.0f * fromLittleEndian16(sens_reg_ptr[2]);  // ASSE Z, va invertito di
                                                      // segno per allineare la
                                                      // terna con il resto del
                                                      // chip.

    /*for(int i=7;i<10;i++)
    {
        if(rawdata<0)
    }
    */

    // converto tutto in microtesla
    for (int i = 7; i < 10; i++)
    {
        data_out[i] =
            rawdata[i] * AK8963_sensitivity;  // conversione in microtesla
        data_out[i] = data_out[i] * (((mag_adj[i - 7] - 128) * 0.5) / 128.0 +
                                     1);  // aggiustamento della scala con i
                                          // parametri di calibrazioni interni
    }

    // Sensitivity adjustment

    magCalibration();  // riaggiusto i dati del magnetometro con la matrice di
                       // calibrazione
    return true;
}

/*
 * Chiamata per aggiornare i valori del giroscopio e dell'accelerometro
 */
bool MPU9255::onSimpleUpdate()
{
    if (initialized)
    {
        uint8_t sens_reg[NUM_RAW_UINT8];
        const uint16_t *sens_reg_ptr = (const uint16_t *)sens_reg;

        masterRead(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_ACCEL_XOUT_H, sens_reg,
                   NUM_RAW_UINT8);

        // AXEL
        for (int i = 0; i < 3; i++)
            rawdata[i] =
                fromBigEndian16(sens_reg_ptr[i]) - rawAxelOffset[i];  // ASSE X

        // TEMP
        rawdata[3] = fromBigEndian16(sens_reg_ptr[3]);  // TEMP
        // GYRO --> elimino gli ultimi 4 bit, tanto sono solo di rumore
        for (int i = 4; i <= 6; i++)
        {
            rawdata[i] =
                fromBigEndian16(sens_reg_ptr[i]) - rawGyroOffset[i - 4];
            if (abs(rawdata[i]) <
                (gyro_min_value / getGyroFSR() *
                 (float)
                     sensor_fsr))  // converting the degrees per second in bits
                rawdata[i] = 0;
        }

        // VALORI OUT ACCELEROMETRO
        for (int i      = 0; i <= 2; i++)
            data_out[i] = (float)rawdata[i] / (float)sensor_fsr * getAxelFSR();

        // FIXME: ricontrollare temperatura
        data_out[3] =
            (((float)rawdata[3] - room_temp_offset) / temp_sensitivity) + 21;

        for (int i      = 4; i <= 6; i++)
            data_out[i] = (float)rawdata[i] / (float)sensor_fsr * getGyroFSR();
        // VALORI DEL MAGNETOMETRO

        if (updateMagValues())  // la funzione torna vero se il magnetometro non
                                // è
                                // saturato, falso se ha saturato
        {
            // AGGIORNO I DATI PER IL CALCOLO DELL'ASSETTO CON IL MAGNETOMETRO
            assetto->update(data_out[4], data_out[5], data_out[6], data_out[0],
                            data_out[1], data_out[2], data_out[7], data_out[8],
                            data_out[9]);
            compassHeading();  // aggiorno anche la bussola
        }
        else
        {
            assetto->updateIMU(data_out[4], data_out[5], data_out[6],
                               data_out[0], data_out[1], data_out[2]);

            configPassThrough();  // ri-invio l'inizializzazione al
                                  // magnetometro,
            // perchè protrebbe anche aver avuto una perdita
            // di configurazione (tipo se traballa il
            // connettore)
            configMagMode(MPU_AK89XX_CONT2_MODE);  // reinvio anche la modalità
                                                   // di lettura.
        }

        data.accel = Vec3(data_out[0], data_out[1], data_out[2]);

        data.gyro = Vec3(data_out[4], data_out[5], data_out[6]);

        data.magnet = Vec3(data_out[7], data_out[8], data_out[9]);

        data.yaw     = assetto->getYaw();
        data.pitch   = assetto->getPitch();
        data.roll    = assetto->getRoll();
        data.compass = compass_heading;

        data.timestamp = getTick();

        return true;
    }
    return false;
}

void MPU9255::compassHeading()
{
    float xh, yh, ayf, axf;
    ayf = data_out[1] / 57.0;  // Convert to rad
    axf = data_out[0] / 57.0;  // Convert to rad
    xh  = data_out[7] * cos(ayf) + data_out[8] * sin(ayf) * sin(axf) -
         data_out[9] * cos(axf) * sin(ayf);
    yh = data_out[8] * cos(axf) + data_out[9] * sin(axf);

    compass_heading =
        atan2((double)yh, (double)xh) * (180 / M_PI) - 90;  // angle in degrees

    if (compass_heading > 0)
    {
        compass_heading = compass_heading - 360;
    }
    compass_heading = 360 + compass_heading;

    // compass_heading = fmod(180.0/M_PI * atan2f(data_out[8],data_out[7]) +
    // 450,360);
}

/*
 ******************************************************************************
 * GETTER METHODS
 ******************************************************************************
 */

std::string MPU9255::getString()
{
    std::stringstream ss;
    ss << std::fixed << std::setprecision(2) << std::setw(7) << std::showpos;

    ss << "\nxA: " << data_out[0] << "\tyA: " << data_out[1]
       << "\tzA: " << data_out[2] << "      ";
    ss << "\nxG: " << data_out[4] << "\tyG: " << data_out[5]
       << "\tzG: " << data_out[6] << "      ";
    ss << "\nxM: " << data_out[7] << "\tyM: " << data_out[8]
       << "\tzM: " << data_out[9] << "      ";

    ss << std::fixed << std::setprecision(0) << std::setw(3)
       << std::setfill('0') << std::noshowpos;
    ss << "\nYAW: " << assetto->getYaw() << "\tPITCH: " << assetto->getPitch()
       << "\tROLL: " << assetto->getRoll() << "      \n"
       << "COMPASS: " << compass_heading << "      ";
    if (gyro_cal_failed)
        return "CALIBRATION FAILED:YOU MOVED THE SENSOR!!\nRESET AND RETRY!\n ";
    else
        return ss.str();
}

std::string MPU9255::getAttitudeString()
{
    std::stringstream ss;
    ss << "Orientation: " << assetto->getYaw() << " " << assetto->getPitch()
       << " " << assetto->getRoll() << "\n";
    return ss.str();
}

uint8_t MPU9255::getAxelFSR()
{
    switch (axelFSR)
    {
        case AFS_2G:
            return 2;
        case AFS_4G:
            return 4;
        case AFS_8G:
            return 8;
        case AFS_16G:
            return 16;
        default:
            return 255;
    }
}

uint16_t MPU9255::getGyroFSR()
{
    switch (gyroFSR)
    {
        case GFS_250DPS:
            return 250;
        case GFS_500DPS:
            return 500;
        case GFS_1000DPS:
            return 1000;
        case GFS_2000DPS:
            return 2000;
        default:
            return 1000;
    }
}

float MPU9255::getYaw()
{
    float temp_yaw = assetto->getYaw();
    return temp_yaw;
}

/*****************************************************************************************************
 * PRINTER METHODS
 ******************************************************************************************************/

void MPU9255::resetMagnetometerCalibration()
{
    memset(mag_cal_offset, 0, sizeof(mag_cal_offset));  // azzero tutto
    memset(mag_cal_matrix, 0, sizeof(mag_cal_matrix));  // azzero tutto
    for (int i               = 0; i < 3; i++)  // metto gli 1 sulla diagonale
        mag_cal_matrix[i][i] = 1;
}

void MPU9255::printDataForCalibration()
{
    // i dati devono essere raw, perchè tanto vengono rinormalizzati
    TRACE("Raw:%d,%d,%d,%d,%d,%d,%d,%d,%d\n", (int)(data_out[0] * 1000),
          (int)(data_out[1] * 1000), (int)(data_out[2] * 1000),
          (int)(data_out[4] * 10), (int)(data_out[5] * 10),
          (int)(data_out[6] * 10), (int)(data_out[7] * 10),
          (int)(data_out[8] * 10), (int)(data_out[9] * 10));
    fflush(stdout);
}

void MPU9255::printCalibrationForGUI()
{
    TRACE("Cal1:%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f\n", mag_cal_offset[0],
          mag_cal_offset[1], mag_cal_offset[2], mag_cal_matrix[0][0],
          mag_cal_matrix[0][1], mag_cal_matrix[0][2], mag_cal_matrix[1][0],
          mag_cal_matrix[1][1], mag_cal_matrix[1][2], mag_cal_matrix[2][0],
          mag_cal_matrix[2][1], mag_cal_matrix[2][2]);
}

/*
 ******************************************************************************
 * SETTER METHODS
 ******************************************************************************
 */
void MPU9255::updateSampleFrequency(float new_freq)
{
    sample_freq = new_freq;       // aggiorno il valore in memoria
    assetto->begin(sample_freq);  // aggiorno il valore nella libreria
}

void MPU9255::setNewCalibrationValues()
{
    if (getchar() == 'C')
        scanf("%f %f %f %f %f %f %f %f %f %f %f %f", &mag_cal_offset[0],
              &mag_cal_offset[1], &mag_cal_offset[2], &mag_cal_matrix[0][0],
              &mag_cal_matrix[0][1], &mag_cal_matrix[0][2],
              &mag_cal_matrix[1][0], &mag_cal_matrix[1][1],
              &mag_cal_matrix[1][2], &mag_cal_matrix[2][0],
              &mag_cal_matrix[2][1], &mag_cal_matrix[2][2]);
}

/*
 ******************************************************************************
 * I2C METHODS
 ******************************************************************************
 */

void MPU9255::masterWrite(uint8_t bit7_address, uint8_t reg_address,
                          uint8_t *data, int len)
{
    int i = 0;
    soft_i2c::sendStart();
    soft_i2c::send(bit7_address << 1);
    soft_i2c::send(reg_address);

    for (i = 0; i < len; i++)
        soft_i2c::send(data[i]);

    soft_i2c::sendStop();
}

void MPU9255::masterRead(uint8_t bit7_address, uint8_t reg_address,
                         uint8_t *data, int len)
{
    int i = 0;
    soft_i2c::sendStart();
    soft_i2c::send((bit7_address << 1));
    soft_i2c::send(reg_address);
    soft_i2c::sendRepeatedStart();
    soft_i2c::send((bit7_address << 1) + 1);

    for (i      = 0; i < (len - 1); i++)
        data[i] = soft_i2c::recvWithAck();

    data[len - 1] = soft_i2c::recvWithNack();

    soft_i2c::sendStop();
}

bool MPU9255::i2cReadIDTest()
{
    uint8_t who_am_i = 0x00;
    bool result      = true;
    TRACE("READING I2C AXEL&GYRO SENSOR ID...");
    masterRead(MPU9255_I2C_7BIT_ADDRESS, MPU_REG_WHO_AM_I, &who_am_i, 1);
    if (who_am_i == 0x73)
    {
        TRACE("\nSENSOR ID: 0x%x . OK: MPU9255 AXEL&GYRO DETECTED.", who_am_i);
    }
    else if (who_am_i == 0x00)
    {
        TRACE("\n\nNO SENSOR DETECTED");
        result = false;
    }
    else
    {
        TRACE(
            "\n\nERROR: wrong sensor detected. Value ridden: 0x%x instead of "
            "0x73",
            who_am_i);
        result = false;
    }

    return result;
}

void MPU9255::readIDMag()
{
    uint8_t who_am_i = 0x00;
    TRACE("\nREADING I2C MAGNETOMETER SENSOR ID...");
    masterRead(AK8963_I2C_7BIT_ADDRESS, MPU_AK89XX_REG_WIA, &who_am_i, 1);
    if (who_am_i == 0x48)
        TRACE("\nSENSOR ID: 0x%x . OK: MPU9255 MAGNETOMETER DETECTED.\n",
              who_am_i);
    else if (who_am_i == 0x00)
        TRACE("\n\nNO SENSOR DETECTED\n\n");
    else
        TRACE(
            "\n\nERROR: wrong sensor detected. Value ridden: 0x%x instead of "
            "0x48\n\n",
            who_am_i);
}

/*
 ******************************************************************************
 * DEBUG METHODS
 ******************************************************************************
 */

void MPU9255::i2cTest()
{
    TRACE("test\n");
    fflush(stdout);
    soft_i2c::sendStart();
    soft_i2c::send(0xAB);
    soft_i2c::send(0xAF);
    soft_i2c::sendStop();
}

void MPU9255::printRawValuesTest()
{
    TRACE("%d %d %d %d %d %d %d\n", rawdata[0], rawdata[1], rawdata[2],
          rawdata[3], rawdata[4], rawdata[5], rawdata[6]);
    fflush(stdout);
}

void MPU9255::printRawData1()
{
    TRACE("%d %d %d %d %d %d\n", rawdata[0], rawdata[1], rawdata[2], rawdata[4],
          rawdata[5], rawdata[6]);
    fflush(stdout);
}
