/* Copyright (c) 2017-2018 Skyward Experimental Rocketry
 * Authors: Francesco Tesolin, Matteo Franceschini, Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef MPU9255_H
#define MPU9555_H

#include <interfaces/endianness.h>
#include <limits.h>
#include <math.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <iomanip>
#include <ostream>
#include <sstream>
#include <string>

#include "MPU9255Data.h"
#include "MPU9255_REG_MAP.h"
#include "MahonyAHRS/MahonyAHRS.h"
#include "sensors/Sensor.h"
#include "hwmapping.h"

#define DIM_RAW_DATA \
    10  // the number of values that the sensor contains: 3axel
        // +3gyro+1temperature+3mag
#define NUM_RAW_UINT8 \
    14  // the number of registers to be ridden from the accelerometer+temp+gyro
        // part of the chip to get the correct values. (magnetometer has a
        // different chip )

using namespace miosix;

using soft_i2c = RogallinaBoard::interfaces::i2c::bus;

/**
 * Class to access the MPU9255 IMU.
 *
 * Contains the methods for updating the values and calibrating the
 * magnetometer.
 */
class MPU9255 : public Sensor
{
public:
    /**
     * The constructor expects the sampling frequency.
     * \param sampling frequency, default 5hz
     */
    MPU9255(float imu_freq = 5);
    virtual ~MPU9255() {}
    /**
     * The method that updates the values by sampling the correct registers and
     * writing on the SPI.
     */
    bool onSimpleUpdate() override;
    /**
     * The method that initializes the sensors: every parameter to set is called
     * here (FSR, Filters, ecc) and in its called methodss.
     */
    bool init() override;
    /**
     * TODO: The method that calls the self test procedures.
     * \return true if the test succedeed
     */
    bool selfTest() override { return true; }

    //////Setters
    /**
     * The method changes the sample frequency once the sensor has been
     * initialized. Useful if the sampling frequency changes during the flight.
     * \param the new sampling frequency.
     */
    void updateSampleFrequency(float new_freq);
    /**
     * The method that updates the magnetometer calibration values with those
     * received from the serial interface. It expects a string that starts with
     * $
     */
    void setNewCalibrationValues();
    /**
     * The method that resets the magnetometer calibration: this helps if a new
     * calibration is wanted, to be independent on the previous one,
     */
    void resetMagnetometerCalibration();

    //////Getters
    /**
     * The method that gets the current Yaw as calculated by the Mahony Library.
     * \return the value of the Yaw
     */
    float getYaw();

    /**
     * The method that updates the values by sampling the correct registers and
     * writing on the SPI.
     * \return a formatted string
     */
    std::string getAttitudeString();

    MPU9255Data* getIMUDataPtr() { return &data; }

    //////Printers
    /**
     * The method that prints the accelerometer, gyro and magnetometer in the
     * correct format for calibration.
     */
    void printDataForCalibration();
    /**
     * The method that prints the calibration values for the GUI  magnetometer
     * calibrator, to let it know which are the raw values.
     */
    void printCalibrationForGUI();

private:
    /**Enum that defines the FSRs of the accelerometer */
    enum axelFSRchoice
    {
        AFS_2G  = 0,   // 2 G
        AFS_4G  = 8,   // 4 G
        AFS_8G  = 16,  // 8 G
        AFS_16G = 24   // 16 G
    };

    /**Enum that defines the FSRs of the gyroscope*/
    enum gyroFSRchoice
    {
        GFS_250DPS  = 0,   // 250 deg/s
        GFS_500DPS  = 8,   // 500 deg/s
        GFS_1000DPS = 16,  // 1000 deg/s
        GFS_2000DPS = 24   // 2000 deg/s
    };

    Mahony*
        assetto;  // The attitude calculation result is provided by this class
    float sample_freq;              // The sampling frequency
    int16_t rawdata[DIM_RAW_DATA];  // The values as ridden from the sensor
    float data_out[DIM_RAW_DATA];   // The values in G, DPS e uT, calibrated
    uint8_t mag_adj[3];  // The sensibility adjustment values stored inside the
                         // sensor, used to correct the ridden values.
    axelFSRchoice axelFSR;     // The FSR choice for the accelerometer
    gyroFSRchoice gyroFSR;     // The GSR choice for the gyroscope
    float room_temp_offset;    // Room temperature offset, used to calibrate the
                               // temperature sensor
    float temp_sensitivity;    // The temperature sensor calibrated sensitivity
    int16_t rawGyroOffset[3];  // The gyro offsets, to be subtracted from the
                               // ridden values
    int16_t rawAxelOffset[3];  // The accelerometer offsets, to be substracted
                               // from the ridden values
    float compass_heading;  // Debug variable used by a method that calculates
                            // the heading just based on the magnetometer
    bool gyro_cal_failed = true;  // A debug variable used while calibrating the
                                  // gyroscope, for being able to call the
                                  // updateValues() method without sampling the
                                  // magnetometer.
    bool initialized = false;

    // CALIBRATION VALUES
    const float sensor_fsr = 32768.0f;
    // ACCELEROMETER
    const float axel_x_offset = 0;     // offset in g, it's to be filled by hand
    const float axel_y_offset = 0;     // offset in g, it's to be filled by hand
    const float axel_z_offset = 0.11;  // offset in g, it's to be filled by hand

    // GYROSCOPE
    const float gyro_x_offset = 0;  // default offset in LSB, it will be
                                    // overwritten at every start, with the new
                                    // values.
    const float gyro_y_offset = 0;  // default offset in LSB, it will be
                                    // overwritten at every start, with the new
                                    // values.
    const float gyro_z_offset = 0;  // default offset in LSB, it will be
                                    // overwritten at every start, with the new
                                    // values.
    const float gyro_min_value = 0.2;  // It defines what's the minimum DEGREES
                                       // PER SECOND that we have to consider
                                       // !=0,this helps in erasing the noise.

    // MAGNETOMETER
    const float AK8963_sensitivity = 1.0 / 32760 * 4912;
    // Use skywardMagCal to generate the matrices in situ.
    float mag_cal_offset[3]    = {32.20, -8.35, 48.63};  // X Y Z
    float mag_cal_matrix[3][3] = {{1.023, -0.002, -0.012},
                                  {-0.002, 0.976, -0.009},
                                  {-0.012, -0.009, 1.001}};

    MPU9255Data data;

    // Magnetic Calibration:(2.6% fit error)

    /**
     * The method reads the ID of the magnetometer, to check if it's reachable
     */
    void readIDMag();
    /**
     * The method reads the ID of the accelerometer and gyroscope, to check if
     * it's reachable
     */
    bool i2cReadIDTest();
    /**
     * The method initializes the I2C ports.
     */
    void i2cInit();
    /**
     * The method sets the FSR of the accelerometer and the digital filter
     * \param the accelerometer FSR.
     */
    void axelInit(axelFSRchoice fsr);
    /**
     * The method sets the FSR of the gyroscope and the digital filter
     * \param the gyroscope fsr.
     */
    bool gyroInit(gyroFSRchoice fsr);
    /**
     * The method sets the MPU9255 in passthrough mode, allowing to reach
     * directly the magnetometor from the I2C bus.
     */
    void configPassThrough();
    /**
     * The method sets the magnetometer to a selected mode, sampling at a given
     * sampling rate or allowing to read the fuse memory.
     */
    void configMagMode(uint8_t mag_mode);
    /**
     * The method updates the values stored in memory with the new ones from the
     * sensor, in passthrough mode.
     * \return false if the magnetometer is out of scale.
     */
    bool updateMagValues();
    /**
     * The method calculates the heading just by using accelerometer and
     * magnetometer. It saves the result in the variable compass heading.
     */
    void compassHeading();
    /**
     * The method reads the sensivity adjustment values stored in the
     * magnetometer.
     */
    void readMagAdj();
    /**
     * The method uses the calibration values to adjust the sampled result to
     * the calibrated one.
     */
    void magCalibration();
    /**
     * The method reads which FSR is currently selected and returns the value in
     * "g" corrisponding to the register selection.
     */
    uint8_t getAxelFSR();
    /**
     * The method reads which FSR is currently selected and returns the value in
     * "degrees per second" corrisponding to the register selection.
     */
    uint16_t getGyroFSR();
    /**
     * The method writes to the I2C a certain amount of chars from a vector.
     */
    void masterWrite(
        uint8_t bit7_address, uint8_t reg_address, uint8_t* data,
        int len);  // ogni sensore ha la sua masterWrite e masterRead privata
                   /**
                    * The method reads from the I2C a certain amount of vchars from a vector.
                    */
    void masterRead(uint8_t bit7_address, uint8_t reg_address, uint8_t* data,
                    int len);  // un driver comune non avrebbe senso, date le
                               // differenze di implementazione del protocollo
                               /**
                                * The method is the string generator for the printData() method.
                                */
    std::string getString();

    /////////Debug methods
    void printRawValuesTest();
    void i2cTest();
    void printRawData1();
    void tempInit();
    bool gyroOffsetReg();
};

#endif  // MPU9255_H
