#include "gps.h"
#include <fcntl.h>
#include <cmath>
#include <sstream>
#include <stdexcept>
#include "debug.h"

#define degreesToRadians(x) (M_PI * x / 180.0)
#define radiansToDegrees(x) (x * 180.0 / M_PI)

using std::isnan;

/*parameter look can be =
   GPBOD - Bearing, origin to destination
   GPBWC - Bearing and distance to waypoint, great circle
   GPGGA - Global Positioning System Fix Data
   GPGLL - Geographic position, latitude / longitude
   GPGSA - GPS DOP and active satellites
   GPGSV - GPS Satellites in view
   GPHDT - Heading, True
   GPR00 - List of waypoints in currently active route
   GPRMA - Recommended minimum specific Loran-C data
   GPRMB - Recommended minimum navigation info
   GPRMC - Recommended minimum specific GPS/Transit data
   GPRTE - Routes
   GPTRF - Transit Fix Data
   GPSTN - Multiple Data ID
   GPVBW - Dual Ground / Water Speed
   GPVTG - Track made good and ground speed
   GPWPL - Waypoint location
   GPXTE - Cross-track error, Measured
   GPZDA - Date & Time
        */
/*
Default addresses:
- Write: 0x3C
- Read: 0x3D

*/

/*****************************************************************************************************
 * CONSTRUCTOR
 ******************************************************************************************************/

GPS::GPS(double noise, uint8_t imu_speed, double gps_freq)
{
    filteredGPS =
        alloc_filter_velocity2d(noise);  // definisco quanto pulire il rumore
    lati.update(NAN);
    longi.update(NAN);
    lati.updateFiltered(NAN);
    longi.updateFiltered(NAN);
    gps_callback_time = 1 / gps_freq;  // salvo il rate con cui vengo chiamato,
                                       // ma ricevo una frequenza
    imu_speed_for_heading = imu_speed;
}

GPS::~GPS() {}

/*****************************************************************************************************
 * INITIALIZING METHODS
 ******************************************************************************************************/
bool GPS::init()
{
    TRACE("Initializing GPS\n");

    f_usart = fopen("/dev/gps", "r");  // open gps serial port
    if (f_usart == nullptr)
    {
        TRACE("Cannot open /dev/gps: errno: %d\n", errno);
        D(perror("Error description:"));
        return false;
    }
    else
    {
        TRACE("/dev/gps opened succesfully\n");
    }
    /*size_t n;
    char c;

#warning "THIS IS A BLOCKING CALL, CHECK IF WE CAN REMOVE IT"
    do
    {
        n = fread(&c, sizeof(char), 1, f_usart);
        D(printf("%c", c));
    } while (n > 0);  // svuoto la stdin fino all'ultimo char, in modo da
                      // eliminare i dati vecchi

    int err = ferror(f_usart);
    if (err)
    {
        printf("Error reading from GPS: %d\n", err);
        perror("Error description:");
        return false;
    }*/

    initialized = true;
    return true;
}

int GPS::addGPSstring()
{
    // fgets accetta al massimo GPS_MAX_LEN caratteri da analizzare a ciascun
    // passaggio
    char str_gps[GPS_MAX_LEN];
    int status = 0;
    fgets(str_gps, GPS_MAX_LEN, f_usart);  // bloccante se non ricevo stringhe!!

    int err = ferror(f_usart);
    if (err)
    {
        TRACE("Error reading from GPS: %d\n", err);
        D(perror("Error description:"));
        clearerr(f_usart);
    }

    std::string temp_string = std::string(str_gps);

    // TRACE("Read: %d\n", temp_string.size());

    /*Considero solo le stringhe che mi interessano*/
    if (temp_string.find("GPGGA") != std::string::npos ||
        temp_string.find("GPVTG") != std::string::npos ||
        temp_string.find("GPRMC") != std::string::npos)
    {
        if (temp_string.front() == '$')
        {
            gps_line_buffer.put(temp_string);  // aggiungo il dato al buffer
        }
    }
    return status;
}

/*****************************************************************************************************
 * VALUES UPDATING METHODS
 ******************************************************************************************************/
bool GPS::onSimpleUpdate()
{
    if (initialized)  // se la lista non è vuota
    {
        data.buf_count = gps_line_buffer.count();
        while (!gps_line_buffer.isEmpty())
        {
            std::string line = "";
            try
            {
                line = gps_line_buffer.pop();
            }
            catch (std::range_error& err)
            {
                TRACE("%s\n", err.what());
                continue;
            }

            switch (minmea_sentence_id(line.c_str(), true))
            {
                case MINMEA_SENTENCE_RMC:
                {
                    struct minmea_sentence_rmc frame;
                    if (minmea_parse_rmc(&frame, line.c_str()))
                    {
                        date.update(frame.date);
                    }
                    break;
                }
                case MINMEA_SENTENCE_GGA:
                {
                    struct minmea_sentence_gga frame;
                    float temp_longitude;
                    float temp_latitude;

                    if (minmea_parse_gga(&frame, line.c_str()))
                    {
                        timenow.update(frame.time);

                        fix_type.update(frame.fix_quality);

                        n_sat.update(frame.satellites_tracked);

                        altitude.update(minmea_tofloat(&frame.altitude));

                        temp_latitude  = minmea_tocoord(&frame.latitude);
                        temp_longitude = minmea_tocoord(&frame.longitude);

                        lati.update(temp_latitude);

                        longi.update(temp_longitude);

                        if (!(isnan(temp_latitude)) &&
                            !(isnan(temp_longitude)))  // controllo che ci
                                                       // sia fix
                        // (che non siano NAN)
                        {
                            update_velocity2d(filteredGPS, temp_latitude,
                                              temp_longitude,
                                              gps_callback_time);
                            lati.updateFiltered(getFilteredLatitude());
                            longi.updateFiltered(getFilteredLongitude());
                            speed.updateFiltered(getFilteredSpeed());
                            course.updateFiltered(getFilteredBearing());
                        }
                    }
                    break;
                }
                case MINMEA_SENTENCE_GST:
                {
                    /*  struct minmea_sentence_gst frame;
                      if (minmea_parse_gst(&frame, line.c_str()))
                      {
                          // niente di utile per ora
                      }*/
                    break;
                }
                case MINMEA_SENTENCE_GSV:
                {
                    /*  struct minmea_sentence_gsv frame;
                      if (minmea_parse_gsv(&frame, line.c_str()))
                      {
                          // niente di utile per ora
                      }*/
                    break;
                }
                case MINMEA_SENTENCE_VTG:
                {
                    struct minmea_sentence_vtg frame;
                    if (minmea_parse_vtg(&frame, line.c_str()))
                    {
                        if (fix_type.getValue() > 0)
                            speed.update(minmea_tofloat(&frame.speed_kph));
                        else
                            speed.update(0);  // se non ho GPS devo considerarmi
                                              // fermo per usare la IMU

                        course.update(
                            minmea_tofloat(&frame.true_track_degrees));
                    }
                    break;
                }
                default:
                    break;
            }
        }

        // Copy the values into the struct
        data.lat     = lati.getValue();
        data.lon     = longi.getValue();
        data.alt     = altitude.getValue();
        data.bearing = course.getValue();
        data.speed   = speed.getValue();
        data.sats    = n_sat.getValue();
        data.fix     = fix_type.getValue();

        data.filt_lat     = lati.getFiltered();
        data.filt_long    = longi.getFiltered();
        data.filt_bearing = course.getFiltered();
        data.filt_speed   = speed.getFiltered();

        data.time = timenow.getValue();
        data.date = date.getValue();

        data.timestamp = miosix::getTick();
        return true;
    }
    return false;
}

/*****************************************************************************************************
 * GETTER METHODS
 ******************************************************************************************************/

GPS::guidance GPS::getGuidanceValues(float dest_latitude, float dest_longitude)
{
    guidance rogallina_temp;
    // pthread_mutex_lock(&atomic_reader);  // inizio sezione critica

    rogallina_temp.heading_to_target =
        getTargetHeading(dest_latitude, dest_longitude);
    rogallina_temp.distance_to_target =
        getTargetDistance(dest_latitude, dest_longitude);
    rogallina_temp.actual_heading = getFilteredBearing();
    rogallina_temp.speed          = getFilteredSpeed();
    rogallina_temp.gps_fix        = fix_type.getValue();

    // pthread_mutex_unlock(&atomic_reader);  // fine sezione critica
    return rogallina_temp;
}

GPSData* GPS::getGPSDataPtr() { return &data; }

float GPS::getTargetHeading(float dest_latitude, float dest_longitude)
{
    float local_buffer = course.getValue();  // in questo modo se non ho fix
                                             // dico che la mia direzione
                                             // desiderata è quella che sto
                                             // attualmente tenendos
    float fLat = degreesToRadians(lati.getFiltered());
    float fLng = degreesToRadians(longi.getFiltered());
    float tLat = degreesToRadians(dest_latitude);
    float tLng = degreesToRadians(dest_longitude);

    if (fix_type.getValue() > 0)
        local_buffer =
            fmod(radiansToDegrees(
                     atan2(sinf(tLng - fLng) * cosf(tLat),
                           cosf(fLat) * sinf(tLat) -
                               sinf(fLat) * cosf(tLat) * cosf(tLng - fLng))) +
                     360,
                 360);

    return local_buffer;
}

float GPS::getTargetDistance(float dest_latitude, float dest_longitude)
{
    float local_buffer = 0;
    float fLat         = lati.getFiltered();
    float flong        = longi.getFiltered();
    if (fix_type.getValue() > 0)
        local_buffer =
            acos(sinf(dest_latitude * M_PI / 180) * sinf(fLat * M_PI / 180) +
                 cosf(dest_latitude * M_PI / 180) * cosf(fLat * M_PI / 180) *
                     cosf(flong * M_PI / 180 - dest_longitude * M_PI / 180)) *
            6371000;

    return local_buffer;
}

float GPS::getFilteredLatitude()
{
    double lat;
    double dummy;
    get_lat_long(filteredGPS, &lat, &dummy);
    return (float)lat;
}

float GPS::getFilteredLongitude()
{
    double lon;
    double dummy;
    get_lat_long(filteredGPS, &dummy, &lon);
    return (float)lon;
}

float GPS::getFilteredBearing() { return (float)get_bearing(filteredGPS); }

float GPS::getFilteredSpeed()
{
    return (float)(get_mph(filteredGPS) * 1.60934);  // In kph
}
