/* Copyright (c) 2017 Skyward Experimental Rocketry
 * Authors: Francesco Tesolin, Matteo Franceschini
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <math.h>
#include <miosix.h>  //per i mutex
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <iostream>
#include <list>
#include <mutex>
#include <string>
#include "ActiveObject.h"
#include "GpsData.h"
#include "debug.h"
#include "ikalman/gps.h"  //KalmanFilter
#include "nmea/minmea.h"  //NMEA parser
#include "sensors/Sensor.h"
#include "utils/SyncedCircularBuffer.h"

#ifndef GPS_DATA_H
#define GPS_DATA_H

using namespace miosix;

#define BUFFER_DIM \
    30  // lunghezza massima delle stringhe per ciascun campo dati del gps.
#define GPS_MAX_LEN 82  // lunghezza massima della stringa gps

static const uint16_t max_speed =
    150;  // la velocità oltre la quale sicuramente il dato è sbagliato, in kmh

static constexpr uint16_t GPS_LINE_BUFFER_SIZE = 100;

/*****************************************************************************************************
 * SUPPORT CLASSES
 ******************************************************************************************************/
/*
 * See GPSBaseClass for function description
*/
class GpsDate
{
public:
    GpsDate()
    {
        date.day   = 0;
        date.month = 0;
        date.year  = 0;
    }

    void update(minmea_date new_date) { date = new_date; }

    minmea_date getValue() { return date; }

private:
    const uint16_t gps_buffer_strings = 20;
    minmea_date date;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        sprintf(buffer, "%02d - %02d - %02d     ", date.day, date.month,
                date.year);
        return buffer;
    }
};

class GpsTime
{
public:
    GpsTime()
    {
        time.hours   = 0;
        time.minutes = 0;
        time.seconds = 0;
    }

    void update(minmea_time new_time) { time = new_time; }

    minmea_time getValue() { return time; }

    /**
     * Get the time in microseconds
     */
    uint64_t getValueMicroseconds()
    {
        // clang-format off
        return time.hours * 60 * 60 * 1000 * 1000 //Hours to microseconds
        		+ time.minutes * 60 *1000 * 1000  //Minutes to microseconds
				+ time.seconds * 1000 * 1000	  //Seconds to microseconds
				+ time.microseconds;

        // clang-format on
    }

private:
    minmea_time time;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        sprintf(buffer, "GMT %02d:%02d:%02d     ", time.hours, time.minutes,
                time.seconds);
        return buffer;
    }
};

class GpsLatitude
{
public:
    GpsLatitude()
    {
        latitude          = 0;
        filtered_latitude = 0;
    }

    void update(float new_latitude) { latitude = new_latitude; }

    float getValue() { return latitude; }

    void updateFiltered(float new_filtered_latitude)
    {
        filtered_latitude = new_filtered_latitude;
    }

    float getFiltered() { return filtered_latitude; }

private:
    float latitude;
    float filtered_latitude;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        sprintf(buffer, "%#08.7f  ", filtered_latitude);
        return buffer;
    }
};

class GpsLongitude
{
public:
    GpsLongitude()
    {
        longitude          = 0;
        filtered_longitude = 0;
    }

    float getValue() { return longitude; }

    void update(float new_longitude) { longitude = new_longitude; }

    void updateFiltered(float new_filtered_longitude)
    {
        filtered_longitude = new_filtered_longitude;
    }

    float getFiltered() { return filtered_longitude; }

private:
    float longitude;
    float filtered_longitude;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        sprintf(buffer, "%#08.7f     ", filtered_longitude);
        return buffer;
    }
};

class GpsCourse
{
public:
    GpsCourse()
    {
        course          = 0;
        filtered_course = 0;
    }

    void update(float new_course) { course = new_course; }

    float getValue() { return course; }

    void updateFiltered(float new_filtered_course)
    {
        filtered_course = new_filtered_course;
    }

    float getFiltered() { return filtered_course; }

private:
    float course;
    float filtered_course;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        sprintf(buffer, "%#6.2f °    ", filtered_course);
        return buffer;
    }
};

class GpsSpeed
{
public:
    GpsSpeed()
    {
        speed          = 0;
        filtered_speed = 0;
    }

    void update(float new_speed)
    {
        if (new_speed > max_speed)
            speed = 0;
        else
            speed = new_speed;
    }

    float getValue() { return speed; }

    void updateFiltered(float new_filtered_speed)
    {
        if (new_filtered_speed > max_speed)
            filtered_speed = 0;
        else
            filtered_speed = new_filtered_speed;
    }

    float getFiltered() { return filtered_speed; }

private:
    float speed;
    float filtered_speed;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        sprintf(buffer, "%#5.2f kmh    ", filtered_speed);
        return buffer;
    }
};

class GpsAltitude
{
public:
    GpsAltitude() { altitude = 0; }

    void update(float new_altitude) { altitude = new_altitude; }

    float getValue() { return altitude; }

private:
    float altitude;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        sprintf(buffer, "%#06.2f m    ", altitude);
        return buffer;
    }
};

class GpsNsat
{
public:
    GpsNsat() { n_sat = 0; }

    void update(uint8_t new_n_sat) { n_sat = new_n_sat; }

    uint8_t getValue() { return n_sat; }

private:
    uint8_t n_sat;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        sprintf(buffer, "%02d", n_sat);
        return buffer;
    }
};

class GpsFixType
{
public:
    GpsFixType() { fix_type = 0; }

    void update(uint8_t new_fix_type) { fix_type = new_fix_type; }

    uint8_t getValue() { return fix_type; }

private:
    uint8_t fix_type;

    std::string getString()
    {
        char buffer[BUFFER_DIM];
        switch (fix_type)
        {
            case 0:
                sprintf(buffer, "%s", "INVALID FIX");
                break;
            case 1:
                sprintf(buffer, "%s", "GPS FIX          ");
                break;
            case 2:
                sprintf(buffer, "%s", "AGPS FIX         ");
                break;
        }

        return buffer;
    }
};

/*****************************************************************************************************
 * GPS CLASS
 ******************************************************************************************************/

/**
 * Class to access the GPS.
 *
 * MODE                    NMEA
 * baudrate                115200
 * sampling rate           5hz
 *
 */
class GPS : public Sensor, public ActiveObject
{
public:
    /**
     * The constructor expects the heaviness of the filtering, default 1, the
     * higher, the heviest.
     * It also expects the rate at which the updateValues() function is being
     * called.
     * \param noise filtering heaviness
     * \param the speed on which the course (heading) is to be considered not
     * reliable if given by the GPS
     * \param sampling frequency, default 5hz
     */
    GPS(double noise = 1, uint8_t imu_speed = 10, double sample_freq = 5);

    /**
    *   Destructor
    */
    virtual ~GPS();

    typedef struct guidance_values  // the struct that contains the new values
                                    // on which we can run the control algorithm
    {
        float heading_to_target;
        float actual_heading;
        float distance_to_target;
        float speed;
        uint8_t gps_fix;

        bool hasFix() const { return gps_fix > 0; }
    } guidance;

    /**
     *  TODO: Calling this method flushes the stdin, to make sure that only
     * updated data is present.
     */
    bool init() override;

    /**
     * TODO: this method tests if the GPS is correctly connected
     * \return TRUE if it's ok, FALSE if not connected
     */
    bool selfTest() override { return true; }

    /**
     * Calling this method addes the latest gps_string to the local list.
     * WARNING: due to the fact that uses the puts function, it's blocking!!!!
     */
    int addGPSstring();
    /**
     * Calling this method parses the new strings from the gps strings list.
     *
     */
    bool onSimpleUpdate() override;

    /**
     * Calling this method gets the new guidance values for the wing.
     * This method is atomic, as when called it locks the parsing of new data to
     * prevent incomplete readings.
     */
    guidance getGuidanceValues(float dest_latitude, float dest_longitude);

    GPSData* getGPSDataPtr();

    void stop() { stop_driver = true; }

protected:
    void run() override
    {
        while (true)
        {
            addGPSstring();  // This blocks until it receives a new line on the
                             // gps USART
            if (stop_driver)
                return;
        }
    }

private:
    // pthread_mutex_t atomic_reader    = PTHREAD_MUTEX_INITIALIZER;
    // pthread_mutex_t atomic_list_push = PTHREAD_MUTEX_INITIALIZER;

    double gps_callback_time;  // stores the callback time of the function that
    uint8_t imu_speed_for_heading;
    SyncedCircularBuffer<std::string, GPS_LINE_BUFFER_SIZE> gps_line_buffer;
    GpsDate date;        // stores the date
    GpsTime timenow;     // stores the time in GMT+2
    GpsLatitude lati;    // stores the latitude
    GpsLongitude longi;  // stores the longitude
    GpsCourse course;  // stores the heading, as given by the GPS and the filter
    GpsSpeed speed;    // stores the speed
    GpsAltitude altitude;  // stores the altitude
    GpsNsat n_sat;         // stores the salellites in view
    GpsFixType fix_type;   // stores if there's fix or not

    bool stop_driver = false;

    GPSData data;

    KalmanFilter filteredGPS;

    bool initialized = false;

    FILE* f_usart = nullptr;

    /**
     * Calling this method returns the angle between the current position and
     * the destination position
     * \return Float with the heading to the target
     */
    float getTargetHeading(float dest_latitude, float dest_longitude);

    /**
    * Calling this method returns the distance between the current position and
    * the destination position, using the spherical law approximation
    * \return Float with the distance from target
    */
    float getTargetDistance(float dest_latitude, float dest_longitude);

    /**
    * Calling this method returns the filtered latitude with a 2D kalman filter
    * \return the latitude in degrees, filtered with the kalman filter
    */
    float getFilteredLatitude();

    /**
    * Calling this method returns the filtered longitude with a 2D kalman filter
    * \return the longitude in degrees, filtered with the kalman filter
    */
    float getFilteredLongitude();

    /**
     * Calling this method returns the filtered bearing with a 2D kalman filter
     * \return the current heading, filtered with the kalman filter
     */
    float getFilteredBearing();

    /**
     * Calling this method returns the filtered speed with a 2D kalman filter
     * \return the current speed in kmh, filtered with the kalman filter
     */
    float getFilteredSpeed();
};

#endif
