/* Copyright (c) 2015-2018 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef SRC_SHARED_BOARDS_HELITEST_SYSTEMDIAGNOSTICS_SYSTEMDIAGNOSTICS_H
#define SRC_SHARED_BOARDS_HELITEST_SYSTEMDIAGNOSTICS_SYSTEMDIAGNOSTICS_H

#include <map>

#include <diagnostic/CpuMeter.h>
#include <logger/Logger.h>
#include <miosix.h>
#include <Singleton.h>

#include "SystemData.h"


using std::map;

using miosix::FastMutex;
using miosix::Lock;

using miosix::MemoryProfiling;
/**
 * Class used to store data about the various threads, such as current and min
 * stack.
 */
class SystemDiagnostics : public Singleton<SystemDiagnostics>
{
    friend class Singleton<SystemDiagnostics>;

public:
    /**
     * Stores the thread diagnostics and logs them.
     * @param d
     */
    void updateThreadDiagnostics(const StackData& d)
    {
        logger.log(d);

        Lock<FastMutex> l(mtxMap);
        threadData[d.thread_id] = d;
    }

    /**
     * Returns the thread with the worst absolute free stack between all the
     * threads.
     * @return ThreadDiagnostics struct
     */
    StackData getWorstStack()
    {
        if (threadData.empty())
        {
            return StackData{THREAD_INVALID, 0, 0, 0};
        }

        Lock<FastMutex> l(mtxMap);

        StackData& worst = threadData.begin()->second;

        for (auto it = threadData.begin(); it != threadData.end(); it++)
        {
            if (it->second.abs_free_stack < worst.abs_free_stack)
            {
                worst = it->second;
            }
        }

        return worst;
    }

    float getCpuUsage() { return averageCpuUtilization(); }

    unsigned int getAbsoluteFreeHeap()
    {
        return MemoryProfiling::getAbsoluteFreeHeap();
    }

    void logSystemData()
    {
        SystemData d{miosix::getTick(), getCpuUsage(),
                     MemoryProfiling::getCurrentFreeHeap(),
                     MemoryProfiling::getAbsoluteFreeHeap()};

        logger.log(d);
    };

    map<uint8_t, StackData> getThreadDiagnostics()
    {
        Lock<FastMutex> l(mtxMap);
        return threadData;
    }

private:
    SystemDiagnostics() {}

    map<uint8_t, StackData> threadData;

    FastMutex mtxMap;
    Logger& logger = Logger::instance();
};

#define sSystemDiagnostics SystemDiagnostics::getInstance()
#define sUpdateDiagnostics \
    SystemDiagnostics::getInstance()->updateThreadDiagnostics

#endif /* SRC_SHARED_BOARDS_HELITEST_SYSTEMDIAGNOSTICS_SYSTEMDIAGNOSTICS_H */
