/*
 * Copyright (c) 2019 Skyward Experimental Rocketry
 * Authors: Luca Erbetta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifdef STANDALONE_CATCH1_TEST
#include "catch-tests-entry.cpp"
#endif

#include <functional>

#include <Singleton.h>
#include <miosix.h>
#include <utils/catch.hpp>

#include "ButtonHandler.h"

using namespace miosix;
using std::bind;

class ButtonMock : public Singleton<ButtonMock>
{
    friend class Singleton<ButtonMock>;

public:
    // Mocked methods
    static void mode(miosix::Mode::Mode_ mode)
    {
        getInstance()->input_mode = mode == Mode::INPUT;
    }

    static int value()
    {
        ButtonMock* instance = getInstance();
        if (instance->input_mode)
            return getInstance()->btn_value;

        TRACE("Returning 0\n");
        return 0;
    }

    // Test methods
    void setValue(int value) { btn_value = value; }

    void reset()
    {
        btn_value  = 0;
        input_mode = false;
    }

    void press(unsigned int duration)
    {
        setValue(1);
        Thread::sleep(duration);
        setValue(0);
        Thread::sleep(TICK_LENGTH * 2);
    }

private:
    int btn_value   = 0;
    bool input_mode = false;
};

ButtonMock* btn = ButtonMock::getInstance();

using ButtonHandler_t = ButtonHandler<ButtonMock>;

class ButtonHandlerTestFixture
{

public:
    ButtonHandlerTestFixture() : handler(nullptr) {}

    ~ButtonHandlerTestFixture()
    {
        if (handler)
        {
            handler->stop();
            delete handler;
        }

        btn->reset();
    }

protected:
    void registerOneCallback()
    {
        using namespace std::placeholders;
        ButtonHandler_t::ButtonCallback cb =
            bind(&ButtonHandlerTestFixture::singleCallback, this, _1);
        handler = new ButtonHandler_t(1, cb);
    }

    void registerMultipleCallbacks()
    {
        using namespace std::placeholders;
        ButtonHandler_t::ButtonCallback cb_short =
            bind(&ButtonHandlerTestFixture::shortPressCallback, this, _1);
        ButtonHandler_t::ButtonCallback cb_long =
            bind(&ButtonHandlerTestFixture::longPressCallback, this, _1);
        ButtonHandler_t::ButtonCallback cb_verylong =
            bind(&ButtonHandlerTestFixture::veryLongPressCallback, this, _1);
        handler = new ButtonHandler_t(1, cb_short, cb_long, cb_verylong);
    }

    void singleCallback(ButtonHandler_t::ButtonPress press_type)
    {
        switch (press_type)
        {
            case ButtonHandler_t::ButtonPress::SHORT:
                ++short_press_count;
                break;
            case ButtonHandler_t::ButtonPress::LONG:
                ++long_press_count;
                break;
            case ButtonHandler_t::ButtonPress::VERY_LONG:
                ++very_long_press_count;
                break;
            default:
                break;
        }
    }

    void shortPressCallback(ButtonHandler_t::ButtonPress press_type)
    {
        UNUSED(press_type);
        ++short_press_count;
    }

    void longPressCallback(ButtonHandler_t::ButtonPress press_type)
    {
        UNUSED(press_type);
        ++long_press_count;
    }

    void veryLongPressCallback(ButtonHandler_t::ButtonPress press_type)
    {
        UNUSED(press_type);
        ++very_long_press_count;
    }

    unsigned int short_press_count     = 0;
    unsigned int long_press_count      = 0;
    unsigned int very_long_press_count = 0;
    ButtonHandler_t* handler;
};

TEST_CASE_METHOD(ButtonHandlerTestFixture,
                 "ButtonHandler - Nothing happens if button is not pressed")
{
    registerMultipleCallbacks();
    handler->start();

    Thread::sleep((VERY_LONG_PRESS_TICKS + 1) * TICK_LENGTH);

    REQUIRE(short_press_count == 0);
    REQUIRE(long_press_count == 0);
    REQUIRE(very_long_press_count == 0);
}

TEST_CASE_METHOD(ButtonHandlerTestFixture,
                 "ButtonHandler - Handle each button press")
{
    SECTION("Different callbacks")
    {
        registerMultipleCallbacks();
        handler->start();

        btn->press(TICK_LENGTH + 5);

        
        CHECK(long_press_count == 0);
        CHECK(very_long_press_count == 0);
        REQUIRE(short_press_count == 1);

        btn->press(TICK_LENGTH * (LONG_PRESS_TICKS + 1));

        CHECK(short_press_count == 1);
        CHECK(very_long_press_count == 0);
        REQUIRE(long_press_count == 1);
        

        btn->press(TICK_LENGTH * (VERY_LONG_PRESS_TICKS + 1));

        CHECK(short_press_count == 1);
        CHECK(long_press_count == 1);
        REQUIRE(very_long_press_count == 1);
    }

    SECTION("Same callback")
    {
        registerOneCallback();
        handler->start();

        btn->press(TICK_LENGTH + 5);
  
        CHECK(long_press_count == 0);
        CHECK(very_long_press_count == 0);
        REQUIRE(short_press_count == 1);

        btn->press(TICK_LENGTH * (LONG_PRESS_TICKS + 1));

        CHECK(short_press_count == 1); 
        CHECK(very_long_press_count == 0);
        REQUIRE(long_press_count == 1);

        btn->press(TICK_LENGTH * (VERY_LONG_PRESS_TICKS + 1));


        CHECK(short_press_count == 1);
        CHECK(long_press_count == 1);
        REQUIRE(very_long_press_count == 1);
    }
}

TEST_CASE_METHOD(ButtonHandlerTestFixture,
                 "ButtonHandler - Just short of each button press")
{
    SECTION("Different callbacks")
    {
        registerMultipleCallbacks();
        handler->start();
        
        REQUIRE(long_press_count == 0);
        REQUIRE(very_long_press_count == 0);
        REQUIRE(short_press_count == 0);

        btn->press(TICK_LENGTH * (LONG_PRESS_TICKS - 1));

        REQUIRE(short_press_count == 1);
        REQUIRE(very_long_press_count == 0);
        REQUIRE(long_press_count == 0);
        

        btn->press(TICK_LENGTH * (VERY_LONG_PRESS_TICKS - 1));

        REQUIRE(short_press_count == 1);
        REQUIRE(long_press_count == 1);
        REQUIRE(very_long_press_count == 0);
    }

    SECTION("Same callback")
    {
        registerOneCallback();
        handler->start();
   
        REQUIRE(long_press_count == 0);
        REQUIRE(very_long_press_count == 0);
        REQUIRE(short_press_count == 0);

        btn->press(TICK_LENGTH * (LONG_PRESS_TICKS - 1));

        REQUIRE(short_press_count == 1);
        REQUIRE(very_long_press_count == 0);
        REQUIRE(long_press_count == 0);
        

        btn->press(TICK_LENGTH * (VERY_LONG_PRESS_TICKS - 1));

        REQUIRE(short_press_count == 1);
        REQUIRE(long_press_count == 1);
        REQUIRE(very_long_press_count == 0);
    }
}